# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from copy import deepcopy

from numpy import allclose

from .multi_field import MultiField
from .operators.operator import _OpChain, _OpProd, _OpSum
from .operators.simple_linear_operators import FieldAdapter
from .sugar import domain_union, from_random
from .utilities import myassert


def _optimise_operator(op):
    """
    optimises operator trees, so that same operator subtrees are not computed twice.
    Recognizes same subtrees and replaces them at nodes.
    Recognizes same leaves and structures them.
    Works partly inplace, rendering the old operator unusable"""

    # Format: List of tuple [op, parent_index, left=True right=False]
    nodes = []
    # Format: ID: index in nodes[]
    id_dic = {}
    # Format: [parent_index, left]
    leaves = set()

    # helper functions
    def readable_id():
        # Gives out letters to prepend field_adapter ids for cosmetics
        # Start at 'A'
        current_letter = 65
        repeats = 1
        while True:
            yield chr(current_letter)*repeats
            current_letter += 1
            if current_letter == 91:
                # skip specials
                current_letter += 6
            elif current_letter == 123:
                # End at z and start at AA
                current_letter = 65
                repeats += 1

    prepend_id = readable_id()

    def isnode(op):
        return isinstance(op, (_OpSum, _OpProd))


    def left_parser(left_bool):
        return '_op1' if left_bool else '_op2'

    def get_duplicate_keys(k_list, dic):
        for item in list(dic.items()):
            if len(item[1]) > 1:
                k_list.append(item[0])

    # Main algorithm functions
    def rebuild_domains(index):
        """Goes bottom up to fix domains which were destroyed by plugging in field adapters"""
        cond = True
        while cond:
            op = nodes[index][0]
            for attr in ('_op1', '_op2'):
                if isinstance(getattr(op, attr), _OpChain):
                    getattr(op, attr)._domain = getattr(op, attr)._ops[-1].domain
            if isnode(op):
                # Some problems doing this on non-multidomains, because one side becomes a multidomain and the other not
                try:
                    op._domain = domain_union((op._op1.domain, op._op2.domain))
                except AttributeError:
                    import warnings
                    warnings.warn('Operator should be defined on a MultiDomain')
                    pass

            index = nodes[index][1]
            cond = type(index) is int


    def recognize_nodes(op, active_node, left):
        # If nothing added - is a leaf!
        isleaf = True
        if isinstance(op, _OpChain):
           for i in range(len(op._ops)):
                if isnode(op._ops[i]):
                    nodes.append((op._ops[i], active_node, left))
                    isleaf = False
        elif isnode(op):
            nodes.append((op, active_node, left))
            isleaf = False
        if isleaf:
            leaves.add((active_node, left))


    def equal_nodes(op):
        # BFS-Algorithm which fills the nodes list and id_dic dictionary
        # Does not scan equal subtrees multiple times
        list_index_traversed = 0
        recognize_nodes(op, None, None)

        while list_index_traversed < len(nodes):
            # Visit node
            active_node = nodes[list_index_traversed][0]

            # Check whether exists already
            try:
                id_dic[id(active_node)] = id_dic[id(active_node)] + [list_index_traversed]
                match = True
            except KeyError:
                id_dic[id(active_node)] = [list_index_traversed]
                match = False
            # Check vertices for nodes
            if not match:
                recognize_nodes(active_node._op1, list_index_traversed, True)
                recognize_nodes(active_node._op2, list_index_traversed, False)

            list_index_traversed += 1

    edited = set()

    def equal_leaves(leaves):
        id_leaf = {}
        # Find matching leaves
        def write_to_dic(leaf, leaf_op_id):
            try:
                id_leaf[leaf_op_id] = id_leaf[leaf_op_id] + (leaf,)
            except KeyError:
                id_leaf[leaf_op_id] = (leaf,)

        for leaf in leaves:
            parent = nodes[leaf[0]][0]
            attr = left_parser(leaf[1])
            leaf_op = getattr(parent, attr)
            if isinstance(leaf_op, _OpChain):
                leaf_op_id = ''
                for i in reversed(leaf_op._ops):
                    leaf_op_id += str(id(i))
                    if not isinstance(i, FieldAdapter):
                        # Do not optimise leaves which only have equal FieldAdapters
                        write_to_dic(leaf, leaf_op_id)
                        break
            else:
                if not isinstance(leaf_op, FieldAdapter):
                    write_to_dic(leaf, str(id(leaf_op)))


        # Unroll their OpChain and see how far they are equal
        key_list_leaf = []
        same_leaf = {}
        get_duplicate_keys(key_list_leaf, id_leaf)

        for key in key_list_leaf:
            to_compare = []
            for leaf in id_leaf[key]:
                parent = nodes[leaf[0]][0]
                attr = left_parser(leaf[1])
                leaf_op = getattr(parent, attr)
                if isinstance(leaf_op, _OpChain):
                    to_compare.append(tuple(reversed(leaf_op._ops)))
                else:
                    to_compare.append((leaf_op,))
            first_difference = 1
            max_diff = min(len(i) for i in to_compare)
            if not max_diff == 1:
                compare_iterator = iter(to_compare)
                first = next(compare_iterator)
                while all(first[first_difference] == rest[first_difference] for rest in compare_iterator):
                    first_difference += 1
                    if first_difference >= max_diff:
                        break
                    compare_iterator = iter(to_compare)
                    first = next(compare_iterator)

            common_op = to_compare[0][:first_difference]
            res_op = common_op[0]
            for ops in common_op[1:]:
                res_op = ops @ res_op

            same_leaf[key] = [res_op, FieldAdapter(res_op.target, next(prepend_id) + str(id(res_op)))]

            for leaf in id_leaf[key]:
                parent = nodes[leaf[0]][0]
                edited.add(id_dic[id(parent)][0])
                attr = left_parser(leaf[1])
                leaf_op = getattr(parent, attr)
                if isinstance(leaf_op, _OpChain):
                    if first_difference == len(leaf_op._ops):
                        setattr(parent, attr, same_leaf[key][1])
                    else:
                        leaf_op._ops = leaf_op._ops[:-first_difference] + (same_leaf[key][1],)
                else:
                    setattr(parent, attr, same_leaf[key][1])
        return key_list_leaf, same_leaf

    equal_nodes(op)

    key_temp = []
    key_list_op, same_op = equal_leaves(leaves)
    cond = True
    while cond:
        key_temp, same_op_temp = equal_leaves(leaves)
        key_list_op += key_temp
        same_op.update(same_op_temp)
        cond = len(same_op_temp) > 0
    key_temp.clear()

    # Cut subtrees
    key_list_node = []
    key_list_subtrees = []
    same_node = {}
    same_subtrees = {}
    subtree_leaves = set()

    get_duplicate_keys(key_list_node, id_dic)

    for key in key_list_node:
        same_node[key] = [nodes[id_dic[key][0]][0],
                          FieldAdapter(nodes[id_dic[key][0]][0].target, next(prepend_id) + str(key))]

        for node_indices in id_dic[key]:
            edited.add(node_indices)
            parent = nodes[nodes[node_indices][1]][0]
            attr = left_parser(nodes[node_indices][2])
            if isinstance(getattr(parent, attr), _OpChain):
                getattr(parent, attr)._ops = getattr(parent, attr)._ops[:-1] + (same_node[key][1],)
            else:
                setattr(parent, attr, same_node[key][1])
            # Nodes have been replaced - treat replacements now as leaves
            subtree_leaves.add((nodes[node_indices][1], nodes[node_indices][2]))
        cond = True
        while cond:
            key_temp1, same_temp = equal_leaves(subtree_leaves)
            key_temp = key_temp1 + key_temp
            same_subtrees.update(same_temp)
            cond = len(same_temp) > 0
        key_list_subtrees += key_temp + [key, ]
        key_temp.clear()
        subtree_leaves.clear()
    same_subtrees.update(same_node)

    for index in edited:
        rebuild_domains(index)
    if isinstance(op, _OpChain):
        op._domain = op._ops[-1].domain

    # Insert trees before leaves
    for key in key_list_subtrees:
        op = op.partial_insert(same_subtrees[key][1].adjoint(same_subtrees[key][0]))
    for key in reversed(key_list_op):
        op = op.partial_insert(same_op[key][1].adjoint(same_op[key][0]))
    return op





def optimise_operator(op):
    """
    Merges redundant operations in the tree structure of an operator.
    For example it is ensured that for ``f@x + x`` the operator ``x`` is only computed once.
    It is supposed to be used on the whole operator chain before doing minimisation.

    Currently optimises only ``_OpChain``, ``_OpSum`` and ``_OpProd`` and not their linear pendants
    ``ChainOp`` and ``SumOperator``.

    Parameters
    ----------
    op : Operator
        Operator with a tree structure.

    Returns
    -------
    op_optimised : Operator
        Operator with same input/output, but optimised tree structure.

    Notes
    -----
    Operators are compared only by id, so best results are achieved when the following code

    >>> from nifty import UniformOperator, DomainTuple
    >>> uni1 = UniformOperator(DomainTuple.scalar_domain()
    >>> uni2 = UniformOperator(DomainTuple.scalar_domain()
    >>> op = (uni1 + uni2)*(uni1 + uni2)

    is replaced by something comparable to

    >>> uni = UniformOperator(DomainTuple.scalar_domain())
    >>> uni_add = uni + uni
    >>> op = uni_add * uni_add

    After optimisation the operator is as fast as

    >>> op = (2*uni)**2
    """
    op_optimised = deepcopy(op)
    op_optimised = _optimise_operator(op_optimised)
    test_field = from_random(op.domain)
    if isinstance(op(test_field), MultiField):
        for key in op(test_field).keys():
            myassert(allclose(op(test_field).asnumpy()[key], op_optimised(test_field).asnumpy()[key], 1e-10))
    else:
        myassert(allclose(op(test_field).asnumpy(), op_optimised(test_field).asnumpy(), 1e-10))
    return op_optimised



# %%

import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
from jax import random

import nifty.re as jft
from nifty.re.tree_math.util import sqrtm

jax.config.update("jax_enable_x64", True)

seed = 42
key = random.key(seed)

# %% [markdown]
# ## N-dimensional, non-diagonal, variable-covariance Gaussian likelihood
#
# This example demonstrates how to use the `NDVariableCovarianceGaussian` likelihood.
# You can choose between a parametrisation using the covariance matrix and its inverse,
# the precision matrix. We strongly recommend parametrising the matrix in its eigenspace
# and build the rotated covariance as needed (see below for an example).

# %% [markdown]
# ## Create synthetic data

# %%

N = 100
mean = jnp.array([1.0, -1.0])
std = [1.0, 0.5]
corr = 0.75
cov_true = jnp.array(
    [
        [std[0] * std[0], std[0] * std[1] * corr],
        [std[0] * std[1] * corr, std[1] * std[1]],
    ]
)
key, key1 = random.split(key, 2)
random_numbers = random.normal(key1, shape=(N, 2))
data = jnp.einsum("ij,kj->ki", sqrtm(cov_true), random_numbers)
data += mean[None, :]
assert data.shape == (N, 2)

# %% [markdown]
# ## Build likelihood and forward-model

# %%


class Forward(jft.Model):
    def __init__(self):
        self.mean = jft.NormalPrior(0.0, 5.0, name="mean", shape=(1, 2))
        self.std_0 = jft.LogNormalPrior(1.0, 1.0, name="std_0", shape=())
        self.std_1 = jft.LogNormalPrior(1.0, 1.0, name="std_1", shape=())
        self.angle = jft.UniformPrior(0.0, jnp.pi / 2, name="angle", shape=())
        super().__init__(
            init=self.mean.init | self.std_0.init | self.std_1.init | self.angle.init
        )

    def __call__(self, xi):
        mean = self.mean(xi)
        mean = jnp.broadcast_to(mean, (N, 2))
        std_0 = self.std_0(xi)
        std_1 = self.std_1(xi)
        angle = self.angle(xi)
        diag = jnp.array([[std_0, 0], [0, std_1]])
        rot = jnp.array(
            [[jnp.cos(angle), -jnp.sin(angle)], [jnp.sin(angle), jnp.cos(angle)]]
        )
        cov = rot @ diag @ rot.T
        cov = jnp.broadcast_to(cov[None, ...], (N, 2, 2))
        return (mean, cov)


fwd_model = Forward()
lh = jft.NDVariableCovarianceGaussian(data).amend(fwd_model)

# %% [markdown]
# ## Run inference

# %%
key, key2, key3 = random.split(key, 3)
samps, state = jft.optimize_kl(
    lh,
    1.0e-2 * jft.Vector(lh.init(key2)),
    key=key3,
    n_total_iterations=5,
    n_samples=10,
    sample_mode="nonlinear_resample",
    nonlinearly_update_kwargs=dict(
        minimize_kwargs=dict(
            name="SN", xtol=1.0e-4, cg_kwargs=dict(name="SNCG"), maxiter=10
        ),
    ),
)

# %% [markdown]
# ## Plot results

# %%


def plot_mean_cov(mean, cov, lbl, color, alpha, *, ax):
    v, U = jnp.linalg.eigh(cov)
    theta = jnp.arctan2(U[1, 1], U[0, 1])
    v = jnp.sqrt(v)
    ell = plt.matplotlib.patches.Ellipse(
        mean,
        3.0 * v[1],
        3.0 * v[0],
        angle=theta * 180 / jnp.pi,
        edgecolor=color,
        fill=False,
        alpha=alpha,
        label=lbl,
    )
    ax.add_patch(ell)
    ax.plot(mean[0], mean[1], "o", color=color, markersize=2, alpha=alpha)


fig, ax = plt.subplots()
ax.scatter(data[:, 0], data[:, 1], marker=".", s=5, label="data")
# Plot samples
for i, samp in enumerate(samps):
    mean = fwd_model(samp)[0][0]
    cov = fwd_model(samp)[1][0]
    lbl = "Samples 3 sigma" if i == 0 else ""
    plot_mean_cov(mean, cov, lbl, "red", alpha=0.5, ax=ax)
plot_mean_cov(mean, cov_true, "Truth 3 sigma", "k", 1.0, ax=ax)
ax.set_xlabel("x")
ax.set_ylabel("y")
ax.legend()
fig.savefig("results_NDVCG.png", dpi=400)
plt.show()



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
#
# Copyright 2024 Gordian Edenhofer
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS”
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2024 Max-Planck-Society


_config = dict(
    hartley_convention="non_canonical_hartley",
    break_on_device_copy=False,
    fail_on_device_copy=False,
    fail_on_nontrivial_anyarray_creation_on_host=False,
)


def update(key, value, /):
    """Update the global configuration of NIFTy and NIFTy.re

    Parameters
    ----------
    key : str
        Identifier for the configuration option.
    value : Any
        Value for the configuration option.


    Currently, the following configuration options are available:

    - "hartley_convention": one of "non_canonical_hartley" or
      "canonical_hartley" for ducc's old non-canonical Hartley convention
      respectively ducc's new canononical Hartley convention
    """
    global _config
    if not isinstance(key, str):
        raise TypeError(f"key must be a string; got {key!r}")
    key = key.lower()
    if key == "hartley_convention":
        if not isinstance(value, str):
            raise TypeError(f"value to {key!r} must be a string; got {value!r}")
        if value in ("ducc_hartley", "non_canonical_hartley"):
            value = "non_canonical_hartley"
        elif value in ("ducc_fht", "canonical_hartley"):
            value = "canonical_hartley"
        else:
            raise ValueError(f"invalid value to {key!r}; got {value!r}")
    _config[key] = value



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from dataclasses import dataclass
from functools import partial
from typing import Iterable, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
from jax import vmap

# NOTE, jhealpix must be a top-level import so that the global jnp.ndarray defined
# therein are defined at import time instead during jax.jit as this would lead to
# seemingly leaky tracers.
from . import jhealpix
from .grid import Grid, GridAtLevel, MGrid, MGridAtLevel, OpenGrid, OpenGridAtLevel


@dataclass()
class HEALPixGridAtLevel(GridAtLevel):
    nside: int
    nest: bool

    def __init__(
        self,
        shape=None,
        splits=None,
        parent_splits=None,
        *,
        nside: int = None,
        nest=True,
    ):
        if shape is not None:
            assert nside is None
            assert np.ndim(shape) == 1 and shape.size == 1
            nside = (shape[0] / 12) ** 0.5
        if int(nside) != nside:
            raise TypeError(f"invalid nside {nside!r}; expected int")
        if nest is not True:
            raise NotImplementedError("only nested order currently supported")
        if splits is not None:
            splits = np.atleast_1d(splits)
            assert np.ndim(splits) == 1 and splits.size == 1
            if not (splits[0] == 1 or splits[0] % 4 == 0):
                raise AssertionError()
        if parent_splits is not None:
            parent_splits = np.atleast_1d(parent_splits)
            assert np.ndim(parent_splits) == 1 and parent_splits.size == 1
            if not (parent_splits[0] == 1 or parent_splits[0] % 4 == 0):
                raise AssertionError()
        self.nside = int(nside)
        self.nest = nest
        size = 12 * self.nside**2
        super().__init__(shape=size, splits=splits, parent_splits=parent_splits)

    def neighborhood(self, index, window_size: Iterable[int]):
        index = jnp.atleast_1d(index)
        if not isinstance(window_size, int):
            window_size = window_size[0]
        assert index.shape[0] == 1
        dtp = jnp.result_type(index)
        if window_size == 1:
            return index[..., jnp.newaxis]
        if window_size == self.size:
            assert np.all(index >= 0) and np.all(index < self.size)
            nbrs = np.arange(self.size, dtype=dtp)
            nbrs = nbrs[(np.newaxis,) * index.ndim + (slice(None),)]
            return (index[..., jnp.newaxis] + nbrs) % self.size
        if window_size == 9:
            f = partial(jhealpix.get_all_neighbours_valid, self.nside, nest=self.nest)
            for _ in range(index.ndim - 1):
                f = vmap(f)
            nbrs = f(index[0])[jnp.newaxis, ...]
            return jnp.concatenate((index[..., jnp.newaxis], nbrs), axis=-1).astype(dtp)
        nie = "only zero, 1st and all neighbors allowed for now"
        raise NotImplementedError(nie)

    def index2coord(self, index, **kwargs):
        assert index.shape[0] == 1
        f = partial(jhealpix.pix2vec, self.nside, nest=self.nest)
        for _ in range(index.ndim - 1):
            f = vmap(f)
        cc = f(index[0])
        return jnp.stack(cc, axis=0)

    def coord2index(self, coord, dtype=np.uint64, **kwargs):
        assert coord.shape[0] == 3
        f = partial(jhealpix.vec2pix, self.nside, nest=self.nest)
        for _ in range(coord.ndim - 1):
            f = vmap(f)
        idx = f(*(cc for cc in coord))
        return (idx[jnp.newaxis, ...]).astype(dtype)

    def index2volume(self, index, **kwargs):
        r = 1.0
        surface = 4 * np.pi * r**2
        return (surface / self.size)[(np.newaxis,) * index.ndim]


@dataclass()
class HEALPixGrid(Grid):
    def __init__(
        self,
        *,
        nside0: Optional[int] = None,
        nside: Optional[int] = None,
        depth: Optional[int] = None,
        nest=True,
        shape0=None,
        splits=None,
    ):
        """HEALPix pixelization grid.

        The grid can be defined either via depth and exactly one of (nside0, nside, shape0)
        or via nside and one of (nside0, shape0).
        """
        self.nest = nest

        assert shape0 is None or isinstance(shape0, (int, tuple, np.ndarray))
        if shape0 is not None:
            assert nside0 is None
            shape0 = np.asarray(shape0).ravel()
            assert shape0.size == 1, (
                "shape0 must be a scalar or a single-element array/tuple"
            )
            (shape0,) = shape0
            assert isinstance(shape0, int)
            # Check whether the shape is a valid HEALPix shape
            assert shape0 > 0 and shape0 % 12 == 0
            nside0 = (shape0 / 12) ** 0.5
            assert np.isclose(nside0, round(nside0), atol=1.0e-10)
            nside0 = round(nside0)

        assert nside is None or (
            isinstance(nside, int)
            and nside > 0
            and (nside & (nside - 1)) == 0  # power of 2
        )
        assert nside0 is None or (
            isinstance(nside0, int)
            and nside0 > 0
            and (nside0 & (nside0 - 1)) == 0  # power of 2
        )
        assert depth is None or (isinstance(depth, int) and depth >= 0)

        if depth is not None:
            if (nside0 is None) == (nside is None):
                raise ValueError(
                    "Ambiguous initialisation of HEALPixGrid. If depth is given, please supply exactly one of (nside0, nside, shape0)"
                )
            if nside is not None:
                nside0 = nside // 2**depth
        else:
            if (nside is None) or (nside0 is None):
                raise ValueError(
                    "Ambiguous initialisation of HEALPixGrid. If depth is not given, please supply nside and exactly one of (nside0, shape0)"
                )
            assert nside0 <= nside
            depth = np.log2(nside / nside0)
            assert np.isclose(depth, round(depth), atol=1.0e-10)
            depth = round(depth)

        self.nside0 = nside0
        if splits is None:
            splits = (4,) * depth
        super().__init__(
            shape0=12 * self.nside0**2,
            splits=splits,
            atLevel=partial(HEALPixGridAtLevel, nest=self.nest),
        )

    def amend(self, splits=None, *, added_depth: Optional[int] = None):
        if added_depth is not None and splits is not None:
            ve = "only one of `additional_depth` and `splits` allowed"
            raise ValueError(ve)
        if added_depth is not None:
            splits = (4,) * added_depth
        else:
            assert splits is not None
            splits = (splits,) if isinstance(splits, int) else splits
        splits = tuple(np.atleast_1d(s) for s in splits)
        return self.__class__(nside0=self.nside0, splits=self.splits + splits)


class SimpleOpenGridAtLevel(OpenGridAtLevel):
    def __init__(
        self,
        shape,
        splits=None,
        parent_splits=None,
        *,
        shifts0,
        distances0,
        all_splits,
        level=None,
        shifts=None,
        **kwargs,
    ):
        assert level <= len(all_splits)
        shifts = shifts - shifts0 * np.prod(all_splits[:level], axis=0, initial=1.0)
        self.distances = distances0 / np.prod(all_splits[:level], axis=0, initial=1.0)
        super().__init__(
            shape, splits=splits, parent_splits=parent_splits, shifts=shifts, **kwargs
        )

    def index2coord(self, index):
        bc = (slice(None),) + (np.newaxis,) * (index.ndim - 1)
        coord = super().index2coord(index)
        return coord * ((self.shape + 2 * self.shifts) * self.distances)[bc]

    def coord2index(self, coord, dtype=np.uint64):
        bc = (slice(None),) + (np.newaxis,) * (coord.ndim - 1)
        coord = coord / ((self.shape + 2 * self.shifts) * self.distances)[bc]
        return super().coord2index(coord, dtype=dtype)

    def index2volume(self, index):
        vol = super().index2volume(index)
        return vol * np.prod((self.shape + 2 * self.shifts) * self.distances)


def SimpleOpenGrid(
    *,
    min_shape: Tuple[int],
    window_size: Union[int, Tuple[int]] = 3,
    splits: Union[int, Tuple[int], Tuple[Tuple[int]]] = 2,
    distances: Optional[Union[float, Tuple[float]]] = None,
    depth: Optional[int] = None,
    desired_size0: int = 128,
    atLevel: GridAtLevel = SimpleOpenGridAtLevel,
) -> OpenGrid:
    """Create a regular Cartesian grid with a given minimum shape and a default
    volume of unity at the final depth.

    The initialization automatically determines a suitable depth and padding
    (using the `window_size`).

    Amending the grid will increase the resolution while keeping the previous
    grids including their pixel distances the same. Due to padding, the amended
    grids will live on a slightly smaller volume than the previous grids and won't
    start exactly at zero.

    Parameters
    ----------
    min_shape:
        Minimum shape along each dimension. If the shape can be larger than requested
        at no additional costs, it will be automatically extended.
    window_size:
        Size of the refinement window. The larger the window, the more accurate but
        more costly the refinement.
    splits:
        Number of child pixels of a parent pixels.
    distances:
        Distance between pixels if the grid is regular.
    depth:
        Refinement depth.
    desired_size0:
        Approximate size of the initial coarse-most grid.
    """
    min_shape = np.atleast_1d(min_shape)
    if np.ndim(splits) != 2:
        if depth is None:
            desired_shape0 = desired_size0 ** (1.0 / min_shape.ndim)
            desired_shape0 = np.ceil(desired_shape0).astype(np.int_)
            desired_shape0 = np.broadcast_to(desired_shape0, min_shape.shape)
            desired_shape0 = np.min(
                (desired_shape0, (min_shape / 1.5).astype(int)), axis=0
            )
            splits = np.broadcast_to(splits, min_shape.shape)
            depth = max(
                np.emath.logn(splits, min_shape) - np.emath.logn(splits, desired_shape0)
            )
            depth = max(int(np.ceil(depth)), 0)
        splits = np.broadcast_to(splits, (depth,) + min_shape.shape)
    padding = np.ceil((window_size - 1) // 2).astype(np.int_)
    padding = np.broadcast_to(padding, (depth,) + min_shape.shape)

    # Conservative estimate of the shape at zero depth
    shape0 = np.ceil(
        min_shape / np.prod(splits, axis=0, initial=1)
        + (2 + 2 / np.min(splits, axis=0, initial=1))
        * np.max(padding, axis=0, initial=0)
        + 1
    ).astype(np.int_)
    # Exact final shape assuming the above conservative `shape0`
    shape, shifts = shape0, np.zeros_like(shape0, dtype=float)
    for si, pd in zip(splits, padding):
        shape = si * (shape - 2 * pd)
        shifts = si * (shifts + pd)
    shifts0 = shifts / np.prod(splits, axis=0, initial=1)
    distances = 1.0 / shape if distances is None else distances
    distances0 = np.atleast_1d(distances) * np.prod(splits, axis=0, initial=1)
    return OpenGrid(
        shape0=shape0,
        splits=splits,
        padding=padding,
        atLevel=partial(atLevel, shifts0=shifts0, distances0=distances0),
    )


class LogGridAtLevel(SimpleOpenGridAtLevel):
    def __init__(self, *args, coord_offset, coord_scale, **kwargs):
        # NOTE, technically `coord_offset` and `coord_scale` are redundant with
        # `shifts` and `distances`, however, for ease of use, we first let them
        # scale the grid to (0, 1).
        self.coord_offset = coord_offset
        self.coord_scale = coord_scale
        super().__init__(*args, **kwargs)

    @property
    def r_min(self):
        return self.index2coord(np.array([-0.5]))

    @property
    def r_max(self):
        return self.index2coord(np.array([self.shape[0] - 0.5]))

    def index2coord(self, index):
        coord = super().index2coord(index)
        return jnp.exp(self.coord_scale * coord + self.coord_offset)

    def coord2index(self, coord, dtype=np.uint64):
        coord = (jnp.log(coord) - self.coord_offset) / self.coord_scale
        return super().coord2index(coord, dtype=dtype)

    def index2volume(self, index):
        a = (slice(None),) + (np.newaxis,) * index.ndim
        coords = self.index2coord(index + jnp.array([-0.5, 0.5])[a])
        return jnp.prod(coords[1] - coords[0], axis=0, keepdims=True)


def LogGrid(
    *,
    r_min: float,
    r_max: float,
    distances=None,
    **kwargs,
) -> OpenGrid:
    """Logarithmic grid on top of `SimpleOpenGrid` spanning from `r_min` to `r_max`
    at the final depth.
    """
    if distances is not None:
        raise ValueError("`distances` are incompatible with a logarithmic grid")
    if r_min <= 0.0 or r_max <= r_min:
        raise ValueError(f"invalid r_min {r_min!r} or r_max {r_max!r}")
    coord_offset = np.log(r_min)
    coord_scale = np.log(r_max) - coord_offset
    return SimpleOpenGrid(
        **kwargs,
        atLevel=partial(
            LogGridAtLevel, coord_offset=coord_offset, coord_scale=coord_scale
        ),
    )


class HPRadialGridAtLevel(MGridAtLevel):
    def index2coord(self, index, **kwargs):
        coords = super().index2coord(index, **kwargs)
        return coords[:3] * coords[3]

    def coord2index(self, coord, **kwargs):
        assert coord.shape[0] == 3
        r = jnp.linalg.norm(coord, axis=0)[jnp.newaxis, ...]
        coord = jnp.concatenate((coord / r, r), axis=0)
        return super().coord2index(coord, **kwargs)

    def index2volume(self, index):
        grid_hp, grid_r = self.grids
        r_upper = grid_r.index2coord(index[1:2] + 0.5)
        r_lower = grid_r.index2coord(index[1:2] - 0.5)
        A_unity = grid_hp.index2volume(index[0:1])
        return A_unity * (r_upper**3 - r_lower**3) / 3


def HPLogRGrid(
    min_shape: Optional[Tuple[int, int]] = None,
    *,
    nside: Optional[int] = None,
    r_min_shape: Optional[int] = None,
    r_min,
    r_max,
    r_window_size=3,
    nside0=16,
    atLevel=HPRadialGridAtLevel,
) -> MGrid:
    """Meshgrid of a HEALPix grid and a logarithmic grid.

    See `HEALPixGrid` and `LogGrid`."""
    if r_min_shape is None and nside is None:
        hp_size, r_min_shape = min_shape
        nside = (hp_size / 12) ** 0.5
    depth = np.log2(nside / nside0)
    assert depth == int(depth)
    depth = int(depth)
    grid_hp = HEALPixGrid(nside0=nside0, depth=depth)
    grid_r = LogGrid(
        min_shape=r_min_shape,
        r_min=r_min,
        r_max=r_max,
        window_size=r_window_size,
        depth=depth,
    )
    return MGrid(grid_hp, grid_r, atLevel=atLevel)


class BrokenLogGridAtLevel(SimpleOpenGridAtLevel):
    def __init__(
        self,
        *args,
        alpha,
        beta,
        gamma,
        delta,
        epsilon,
        r_min,
        r_linthresh,
        r_max,
        rg_min,
        rg_linthresh,
        rg_max,
        **kwargs,
    ):
        self._alpha = alpha
        self._beta = beta
        self._gamma = gamma
        self._delta = delta
        self._epsilon = epsilon
        self._r_min = r_min
        self._r_linthresh = r_linthresh
        self._r_max = r_max
        self._rg_min = rg_min
        self._rg_linthresh = rg_linthresh
        self._rg_max = rg_max
        super().__init__(*args, **kwargs)

    @property
    def r_min(self):
        return self.index2coord(np.array([-0.5]))

    @property
    def r_max(self):
        return self.index2coord(np.array([self.shape[0] - 0.5]))

    def index2coord(self, index):
        # map to in-between 0 and 1
        coord = super().index2coord(index)
        # map to in-between r_min and r_max
        condlist = [
            coord < self._rg_min,
            (self._rg_min <= coord) & (coord < self._rg_linthresh),
            (self._rg_linthresh <= coord) & (coord < self._rg_max),
            self._rg_max <= coord,
        ]
        funclist = [
            lambda rg: self._gamma / (rg - self._delta),
            lambda rg: self._r_min + self._alpha * (rg - self._rg_min),
            lambda rg: self._r_linthresh
            * jnp.exp(self._beta * (rg - self._rg_linthresh)),
            lambda rg: self._r_max + self._epsilon * (rg - self._rg_max),
        ]
        return jnp.piecewise(coord, condlist, funclist)

    def coord2index(self, coord, dtype=np.uint64):
        # map to in-between 0 and 1
        condlist = [
            coord < self._r_min,
            (self._r_min <= coord) & (coord < self._r_linthresh),
            (self._r_linthresh <= coord) & (coord < self._r_max),
            self._r_max <= coord,
        ]
        funclist = [
            lambda r: self._delta + self._gamma / r,
            lambda r: self._rg_min + (r - self._r_min) / self._alpha,
            lambda r: self._rg_linthresh + jnp.log(r / self._r_linthresh) / self._beta,
            lambda r: self._rg_max + (r - self._r_max) / self._epsilon,
        ]
        coord = jnp.piecewise(coord, condlist, funclist)
        # transform to index
        return super().coord2index(coord, dtype=dtype)

    def index2volume(self, index):
        a = (slice(None),) + (np.newaxis,) * index.ndim
        coords = self.index2coord(index + jnp.array([-0.5, 0.5])[a])
        return jnp.prod(coords[1] - coords[0], axis=0, keepdims=True)


def BrokenLogGrid(
    *,
    r_min: float,
    r_linthresh: float,
    r_max: float,
    distances=None,
    **kwargs,
) -> OpenGrid:
    """Create a broken logarithmic grid on top of `SimpleOpenGrid` spanning from
    `r_min` to `r_max` at the final depth.
    The grid is parametrised by three radii: r_min, r_linthresh, and r_max.
    Between r_min and r_linthresh pixels are spaced linearly (r).
    Between r_linthresh and r_max pixels are spaced logarithmically (exp(r)).

    For available parameters see the `SimpleOpenGrid` docstring in addition the ones below.

    Parameters
    ----------
    r_min:
        Minimum coordinate value.
    r_linthresh:
        Coordinate value at which the grid switches from linear to logarithmic spacing.
    r_max:
        Maximum coordinate value.

    Notes
    -----
    For values below rmin, the (padded) pixels are spaced antilinearly (1/r).
    Above rmax they are spaced linearly (r).
    """
    if distances is not None:
        raise ValueError("`distances` are incompatible with a logarithmic grid")
    if r_min <= 0.0 or r_max <= r_min:
        raise ValueError(f"invalid r_min {r_min!r} or r_max {r_max!r}")
    if r_linthresh < r_min or r_max <= r_linthresh:
        raise ValueError(f"invalid r_0 {r_linthresh!r}")

    # This parametrisation is technically capable of handling a transformation
    # from arbitrary rg_min and rg_max, but in accordance to the LogGrid,
    # we can fix them to 0 and 1 and use the parent class for mapping them there.
    rg_min = 0.0
    rg_max = 1.0
    m = (1.0 - r_min / r_linthresh) / (jnp.log(r_max / r_linthresh))
    rg_linthresh = rg_min / (1 + m) + rg_max * m / (1 + m)
    alpha = r_linthresh / (rg_max - rg_linthresh) * jnp.log(r_max / r_linthresh)
    beta = alpha / r_linthresh
    gamma = -(r_min**2) / alpha
    delta = rg_min + r_min / alpha
    epsilon = r_linthresh * beta * jnp.exp(beta * (rg_max - rg_linthresh))

    return SimpleOpenGrid(
        **kwargs,
        atLevel=partial(
            BrokenLogGridAtLevel,
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            delta=delta,
            epsilon=epsilon,
            r_min=r_min,
            r_linthresh=r_linthresh,
            r_max=r_max,
            rg_min=rg_min,
            rg_linthresh=rg_linthresh,
            rg_max=rg_max,
        ),
    )


def HPBrokenLogRGrid(
    min_shape: Optional[Tuple[int, int]] = None,
    *,
    nside: Optional[int] = None,
    r_min_shape: Optional[int] = None,
    r_min,
    r_linthresh,
    r_max,
    r_window_size=3,
    nside0=16,
    atLevel=HPRadialGridAtLevel,
) -> MGrid:
    """Meshgrid of a HEALPix grid and a broken logarithmic grid.

    See `HEALPixGrid` and `BrokenLogGrid`."""
    if r_min_shape is None and nside is None:
        hp_size, r_min_shape = min_shape
        nside = (hp_size / 12) ** 0.5
    depth = np.log2(nside / nside0)
    assert depth == int(depth)
    depth = int(depth)
    grid_hp = HEALPixGrid(nside0=nside0, depth=depth)
    grid_r = BrokenLogGrid(
        min_shape=r_min_shape,
        r_min=r_min,
        r_linthresh=r_linthresh,
        r_max=r_max,
        window_size=r_window_size,
        depth=depth,
    )
    return MGrid(grid_hp, grid_r, atLevel=atLevel)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np
from scipy.interpolate import CubicSpline
from scipy.stats import beta, gamma, invgamma, laplace, norm

from .. import random
from ..domain_tuple import DomainTuple
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from ..operators.adder import Adder
from ..operators.operator import Operator
from ..sugar import makeOp


def _f_on_np(f, arr):
    fld = Field.from_raw(UnstructuredDomain(arr.shape), arr)
    return f(fld).asnumpy()


class _InterpolationOperator(Operator):
    """
    Calculates a function pointwise on a field by interpolation.
    Can be supplied with a table_func to increase the interpolation accuracy,
    Best results are achieved when table_func(func) is roughly linear.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    func : function
        The function which is applied on the field. Assumed to act on numpy
        arrays.
    xmin : float
        The smallest value for which func will be evaluated.
    xmax : float
        The largest value for which func will be evaluated.
    delta : float
        Distance between sampling points for linear interpolation.
    table_func : function
        Non-linear function applied to table in order to transform the table
        to a more linear space. Assumed to act on `Linearization`s, optional.
    inv_table_func : function
        Inverse of table_func, optional.
    """
    def __init__(self, domain, func, xmin, xmax, delta, table_func=None, inv_table_func=None):
        self._domain = self._target = DomainTuple.make(domain)
        self._xmin, self._xmax = float(xmin), float(xmax)
        self._d = float(delta)
        self._xs = np.arange(xmin, xmax, self._d)
        self._table = func(self._xs)
        if table_func is not None:
            if inv_table_func is None:
                raise ValueError
# MR FIXME: not sure whether we should have this in production code
            a = func(random.current_rng().random(10))
            a1 = _f_on_np(lambda x: inv_table_func(table_func(x)), a)
            np.testing.assert_allclose(a, a1)
            self._table = _f_on_np(table_func, self._table)
        self._inv_table_func = inv_table_func
        self._finalize_init()

    def _finalize_init(self):
        if isinstance(self._xs, np.ndarray) and isinstance(self._table, np.ndarray):
            self._interpolator = CubicSpline(self._xs, self._table)
        # elif device_available():
        #     import cupy
        #     from cupyx.scipy.interpolate import CubicSpline as CuCubicSpline
        #     if isinstance(self._xs, cupy.ndarray) and isinstance(self._table, cupy.ndarray):
        #         self._interpolator = CuCubicSpline(self._xs, self._table)
        #     else:
        #         raise RuntimeError
        else:
            raise RuntimeError
        self._deriv = self._interpolator.derivative()

    # def _device_preparation(self, x):
    #     # Scipy cannot work directly with AnyArrays, therefore store np or cupy arrays
    #     self._xs = AnyArray(self._xs).at(x.device_id)._val
    #     self._table = AnyArray(self._table).at(x.device_id)._val
    #     self._finalize_init()

    def apply(self, x):
        # TODO: Check if new Cupy version with CubicSpline has been released
        self._check_input(x)
        device_id = x.device_id  # Temporary
        x = x.at(-1)  # Temporary
        # self._device_preparation(x)
        lin = x.jac is not None
        xval = x.val.raw if lin else x.raw
        res = self._interpolator(xval)
        res = Field(self._domain, res)
        if lin:
            res = x.new(res, makeOp(Field(self._domain, self._deriv(xval))))
        if self._inv_table_func is not None:
            res = self._inv_table_func(res)
        res = res.at(device_id)  # Temporary
        return res


class InverseGammaOperator(Operator):
    """Transform a standard normal into an inverse gamma distribution.

    The pdf of the inverse gamma distribution is defined as follows:

    .. math::
        \\frac{q^\\alpha}{\\Gamma(\\alpha)}x^{-\\alpha -1}
        \\exp \\left(-\\frac{q}{x}\\right)

    That means that for large x the pdf falls off like :math:`x^{(-\\alpha -1)}`.
    The mean of the pdf is at :math:`q / (\\alpha - 1)` if :math:`\\alpha > 1`.
    The mode is :math:`q / (\\alpha + 1)`.

    The operator can be initialized by setting either alpha and q or mode and mean.
    In accordance to the statements above the mean must be greater
    than the mode. Otherwise would get alpha < 0 and so no mean would be defined.

    This transformation is implemented as a linear interpolation which maps a
    Gaussian onto an inverse gamma distribution.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    alpha : float
        The alpha-parameter of the inverse-gamma distribution.
    q : float or :class:`nifty.cl.field.Field`
        The q-parameter of the inverse-gamma distribution.
    mode: float
        The mode of the inverse-gamma distribution.
    mean: float
        The mean of the inverse-gamma distribution.
    delta : float
        Distance between sampling points for linear interpolation.
    """
    def __init__(self, domain, alpha=None, q=None, delta=1e-2, mode=None, mean=None):
        self._domain = self._target = DomainTuple.make(domain)

        if alpha is not None and q is not None:
            self._alpha = float(alpha)
            self._q = q if isinstance(q, Field) else float(q)
            self._mode = self._q / (self._alpha + 1)
            if self._alpha > 1:
                self._mean = self._q / (self._alpha - 1)
        elif mean is not None and mode is not None:
            if mean < mode:
                raise ValueError('Mean should be greater than mode, otherwise alpha < 0')
            self._mean = float(mean)
            self._mode = float(mode)
            self._alpha = 2 / (self._mean / self._mode - 1) + 1
            self._q = self._mode * (self._alpha + 1)
        else:
            raise ValueError("Either one pair of arguments (mode, mean or alpha, q) must be given.")

        self._delta = float(delta)
        op = _InterpolationOperator(self._domain, lambda x: invgamma.ppf(norm._cdf(x), float(self._alpha)),
                                    -8.2, 8.2, self._delta, lambda x: x.ptw("log"), lambda x: x.ptw("exp"))
        if np.isscalar(self._q):
            op = op.scale(self._q)
        else:
            op = makeOp(self._q) @ op
        self._op = op

    def apply(self, x):
        return self._op(x)

    @property
    def alpha(self):
        """float : The value of the alpha-parameter of the inverse-gamma distribution"""
        return self._alpha

    @property
    def q(self):
        """float : The value of q-parameters of the inverse-gamma distribution"""
        return self._q

    @property
    def mode(self):
        """float : The value of the mode of the inverse-gamma distribution"""
        return self._mode

    @property
    def mean(self):
        """float : The value of the mean of the inverse-gamma distribution. Only existing for alpha > 1."""
        if self._alpha <= 1:
            raise ValueError('mean only existing for alpha > 1')
        return self._mean

    @property
    def var(self):
        """float : The value of the variance of the inverse-gamma distribution. Only existing for alpha > 2."""
        if self._alpha <= 2:
            raise ValueError('variance only existing for alpha > 2')
        return self._q**2 / ((self._alpha - 1)**2 * (self._alpha - 2))


class GammaOperator(Operator):
    """Transform a standard normal into a gamma distribution.

    The pdf of the gamma distribution is defined as follows:

    .. math::
        \\frac{1}{\\Gamma(k)\\theta^k} x^{k-1}
        \\exp \\left( -\\frac{x}{\\theta} \\right)

    This transformation is implemented as a linear interpolation which maps a
    Gaussian onto a gamma distribution.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    alpha : float
        The shape parameter of the gamma distribution.
    beta : float or :class:`nifty.cl.field.Field`
        The rate parameter of the gamma distribution.
    theta : float or :class:`nifty.cl.field.Field`
        The scale parameter of the gamma distribution.
    mean : float
        Mean of the gamma distribution.
    var : float
        Variance of the gamma distribution.
    delta : float
        Distance between sampling points for linear interpolation.
    """
    def __init__(self, domain, alpha=None, beta=None, theta=None, mean=None, var=None, delta=1e-2):
        self._domain = self._target = DomainTuple.make(domain)

        if mean is not None and var is not None:
            theta = var / mean
            alpha = mean / theta

        if beta is None and theta is None:
            raise ValueError("Either beta or theta need to be defined")
        if beta is not None:
            theta = 1 / beta
        self._alpha = float(alpha)
        self._theta = theta if isinstance(theta, Field) else float(theta)

        self._delta = float(delta)
        op = _InterpolationOperator(self._domain, lambda x: gamma.ppf(norm._cdf(x), self._alpha),
                                    -8.2, 8.2, self._delta)
        if np.isscalar(self._theta):
            op = op.scale(self._theta)
        else:
            op = makeOp(self._theta) @ op
        self._op = op

    def apply(self, x):
        return self._op(x)

    @property
    def alpha(self):
        """float : The value of the shape-parameter of the gamma distribution"""
        return self._alpha

    @property
    def beta(self):
        """float : The value of rate parameter of the gamma distribution"""
        return 1 / self._theta

    @property
    def theta(self):
        """float : The value of scale parameter of the gamma distribution"""
        return self._theta

    @property
    def mode(self):
        """float : The value of the mode of the gamma distribution. Exists for alpha >= 1 or k >= 1."""
        if self._alpha < 1:
            raise ValueError('mode only existing for alpha >= 1 or k >= 1')
        return (self._alpha - 1) * self._theta

    @property
    def mean(self):
        """float : The value of the mean of the gamma distribution."""
        return self._alpha * self._theta

    @property
    def var(self):
        """float : The value of the variance of the gamma distribution."""
        return self._alpha * self._theta**2


def LogInverseGammaOperator(domain, alpha, q, delta=1e-2):
    """Transform a standard normal into the log of an inverse gamma distribution.

    See also
    --------
    :class:`InverseGammaOperator`
    """
    op = _InterpolationOperator(domain, lambda x: np.log(invgamma.ppf(norm._cdf(x), float(alpha))),
                                -8.2, 8.2, delta)
    q = np.log(q) if np.isscalar(q) else q.log()
    return Adder(q, domain=op.target) @ op


def BetaOperator(domain, a, b, delta=1e-2):
    """Transform a standard normal into a beta distribution.
    """
    func = lambda x: beta.ppf(norm._cdf(x), a=float(a), b=float(b))
    return _InterpolationOperator(domain, func, -8.2, 8.2, delta)


class UniformOperator(Operator):
    """Transform a standard normal into a uniform distribution.

    The uniform distribution's support is ``[loc, loc + scale]``.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    loc: float
    scale: float

    """
    def __init__(self, domain, loc=0, scale=1):
        self._target = self._domain = DomainTuple.make(domain)
        self._loc = float(loc)
        self._scale = float(scale)

    def apply(self, x):
        self._check_input(x)
        lin = x.jac is not None
        xval = x.val.val if lin else x.val
        res = Field(self._target, self._scale*norm._cdf(xval) + self._loc)
        if not lin:
            return res
        jac = makeOp(Field(self._domain, norm._pdf(xval)*self._scale))
        return x.new(res, jac)

    def inverse(self, field):
        res = norm._ppf((field.val - self._loc) / self._scale)
        return Field(field.domain, res)


class LaplaceOperator(Operator):
    """Transform a standard normal to a Laplace distribution.

    Parameters
    -----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    loc : float

    scale : float

    Note
    ----
    If this operator is called on inputs stored on device, it will copy them to
    host, perform the operations with the CPU and copy back. This is done since
    Cupy does not support stats.laplace yet.
    """
    def __init__(self, domain, loc=0, scale=1):
        self._target = self._domain = DomainTuple.make(domain)
        self._loc = float(loc)
        self._scale = float(scale)

    def apply(self, x):
        # TODO: Check if cupy supports stats.laplace
        self._check_input(x)
        lin = x.jac is not None
        xval = x.val.asnumpy() if lin else x.asnumpy()
        res = Field(self._target, laplace.ppf(norm._cdf(xval), self._loc, self._scale))
        res = res.at(x.device_id)
        if not lin:
            return res
        y = norm._cdf(xval)
        y = self._scale * np.where(y > 0.5, 1/(1-y), 1/y)
        jac = makeOp(Field(self.domain, y*norm._pdf(xval)).at(x.device_id))
        return x.new(res, jac)

    def inverse(self, x):
        res = laplace.cdf(x.asnumpy(), self._loc, self._scale)
        res = norm._ppf(res)
        return Field(x.domain, res).at(x.device_id)



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import operator
from dataclasses import field
from functools import reduce
from typing import Any, Callable, Tuple, TypeVar

import jax
from jax.tree_util import (
    Partial,
    tree_flatten,
    tree_leaves,
    tree_map,
    tree_structure,
    tree_unflatten,
)

from .misc import is_iterable_of_non_iterables, isiterable
from .model import LazyModel, NoValue
from .tree_math import ShapeWithDtype, Vector, conj, has_arithmetics, zeros_like

Q = TypeVar("Q")
P = TypeVar("P")


def _functional_conj(func):
    def func_conj(*args, **kwargs):
        # func^*(x) = (func(x^*))^*
        return conj(func(*conj(args), **conj(kwargs)))

    return func_conj


def _parse_point_estimates(point_estimates, primals):
    if isinstance(point_estimates, (tuple, list)):
        if not isinstance(primals, (Vector, dict)):
            te = "tuple-shortcut point-estimate only availble for dict/Vector "
            te += "type primals"
            raise TypeError(te)
        pe = tree_map(lambda x: False, primals)
        pe = pe.tree if isinstance(primals, Vector) else pe
        for k in point_estimates:
            pe[k] = True
        point_estimates = Vector(pe) if isinstance(primals, Vector) else pe
    if tree_structure(primals) != tree_structure(point_estimates):
        print(primals)
        print(point_estimates)
        te = "`primals` and `point_estimates` pytree structre do no match"
        raise TypeError(te)

    primals_liquid, primals_frozen = [], []
    for p, ep in zip(tree_leaves(primals), tree_leaves(point_estimates)):
        if ep:
            primals_frozen.append(p)
        else:
            primals_liquid.append(p)
    primals_liquid = Vector(tuple(primals_liquid))
    primals_frozen = tuple(primals_frozen)
    return point_estimates, primals_liquid, primals_frozen


def _partial_argument(call, insert_axes, flat_fill):
    """For every non-None value in `insert_axes`, amend the value of `flat_fill`
    at the same position to the argument. Both `insert_axes` and `flat_fill` are
    w.r.t. the whole input argument tuple `arg` of `call(*args)`.
    """
    if not flat_fill and not insert_axes:
        return call

    if len(insert_axes) != len(flat_fill):
        ve = "`insert_axes` and `flat_fill` must be of equal length"
        raise ValueError(ve)
    for iae, ffe in zip(insert_axes, flat_fill):
        if iae is not None and ffe is not None:
            if not isinstance(ffe, (tuple, list)):
                te = (
                    f"`flat_fill` must be a tuple of flattened pytrees;"
                    f" got '{flat_fill!r}'"
                )
                raise TypeError(te)
            iae_leaves = tree_leaves(iae)
            if not all(isinstance(e, bool) for e in iae_leaves):
                te = "leaves of `insert_axes` elements must all be boolean"
                raise TypeError(te)
            if sum(iae_leaves) != len(ffe):
                ve = "more inserts in `insert_axes` than elements in `flat_fill`"
                raise ValueError(ve)
        elif iae is not None or ffe is not None:
            ve = (
                "both `insert_axes` and `flat_fill` must be `None` at the same"
                " positions"
            )
            raise ValueError(ve)
    # NOTE, `tree_flatten` replaces `None`s with list of zero length
    insert_axes, in_axes_td = zip(*(tree_flatten(ia) for ia in insert_axes))

    def insert(*x):
        y = []
        assert len(x) == len(insert_axes) == len(flat_fill) == len(in_axes_td)
        for xe, iae, ffe, iatde in zip(x, insert_axes, flat_fill, in_axes_td):
            if ffe is None and not iae:
                y.append(xe)
                continue
            assert iae and ffe is not None
            assert sum(iae) == len(ffe)
            xe, ffe = list(tree_leaves(xe)), list(ffe)
            ye = [xe.pop(0) if not cond else ffe.pop(0) for cond in iae]
            # for cond in iae:
            #     ye.append(xe.pop(0) if not cond else ffe.pop(0))
            y.append(tree_unflatten(iatde, ye))
        return tuple(y)

    def partially_inserted_call(*x):
        return call(*insert(*x))

    return partially_inserted_call


def partial_insert_and_remove(
    call, insert_axes, flat_fill, *, remove_axes=(), unflatten=None
):
    """Return a call in which `flat_fill` is inserted into arguments of `call`
    at `inset_axes` and subsequently removed from its output at `remove_axes`.

    This function is best understood by example:

    .. code-block:: python

        def _identity(x):
            return x

        # _identity takes exactly one argument, thus `insert_axes` and `flat_fill`
        # are length one tuples
        _id_part = jpartial(
            _identity,
            insert_axes=(jft.Vector({
                "a": (True, False),
                "b": False
            }), ),
            flat_fill=(("THIS IS input['a'][0]", ), )
        )
        out = _id_part(("THIS IS input['a'][1]", "THIS IS input['b']"))
        assert out == jft.Vector(
            {
                "a": ("THIS IS input['a'][0]", "THIS IS input['a'][1]"),
                "b": "THIS IS input['b']"
            }
        )

    """
    call = _partial_argument(call, insert_axes=insert_axes, flat_fill=flat_fill)

    if not remove_axes:
        return call

    remove_axes = tree_leaves(remove_axes)
    if not all(isinstance(e, bool) for e in remove_axes):
        raise TypeError("leaves of `remove_axes` must all be boolean")

    def remove(x):
        x, y = list(tree_leaves(x)), []
        if tree_structure(x) != tree_structure(remove_axes):
            te = (
                f"`remove_axes` ({tree_structure(remove_axes)!r}) is shaped"
                f" differently than output of `call` ({tree_structure(x)!r})"
            )
            raise TypeError(te)
        for maybe_remove, cond in zip(x, remove_axes):
            if not cond:
                y.append(maybe_remove)
        y = unflatten(tuple(y)) if unflatten is not None else y
        return y

    def partially_removed_call(*x):
        return remove(call(*x))

    return partially_removed_call


def _parse_swd(shape):
    leaves = tree_leaves(shape)
    if not all(hasattr(e, "shape") and hasattr(e, "dtype") for e in leaves):
        if is_iterable_of_non_iterables(shape) or not isiterable(shape):
            shape = ShapeWithDtype(shape)
        else:
            te = "`lsm_tangents_shapes` of invalid type"
            raise TypeError(te)
    return shape


class Likelihood(LazyModel):
    """Storage class for keeping track of the energy, the associated
    left-square-root of the metric and the metric.

    Properties
    ----------
    energy : callable
        Function evaluating the negative log-likelihood.
    transformation : callable
        Function evaluating the geometric transformation of the likelihood.
    normalized_residual : callable
        Function evaluating the data residual normalized by the standard
        deviation of the likelihood.
    left_sqrt_metric : callable
        Function applying the left-square-root of the metric.
    metric : callable
        Function applying the metric.
    domain : tree-like structure of ShapeWithDtype
        Structure of the input/parameter space.
    lsm_tangents_shape : tree-like structure of ShapeWithDtype
        Structure of the data space.
    """

    _lsm_tan_shp: Any = None

    def __init__(self, *, domain=NoValue, init=NoValue, lsm_tangents_shape=None):
        # NOTE, `lsm_tangents_shape` is not `normalized_residual` applied to
        # `domain` for e.g. models with a learnable covariance
        self._lsm_tan_shp = _parse_swd(lsm_tangents_shape)
        super().__init__(domain=domain, init=init)

    def __call__(self, primals, **primals_kw):
        """Convenience method to access the `energy` method of this instance."""
        return self.energy(primals, **primals_kw)

    def energy(self, primals, **primals_kw):
        """Applies the energy to `primals`.

        Parameters
        ----------
        primals : tree-like structure
            Position at which to evaluate the energy.
        **primals_kw : Any
           Additional arguments passed on to the energy.

        Returns
        -------
        energy : float
            Energy at the position `primals`.
        """
        nie = "`energy` is not implemented"
        raise NotImplementedError(nie)

    def normalized_residual(self, primals, **primals_kw):
        """Applies the normalized_residual to `primals`.

        Parameters
        ----------
        primals : tree-like structure
            Position at which to evaluate the energy.
        **primals_kw : Any
           Additional arguments passed on to the energy.

        Returns
        -------
        normalized_residual : tree-like structure
            Structure of the same type as lsm_tangents_shape for which the
            normalized_residual is computed.
        """
        nie = "`normalized_residual` is not implemented"
        raise NotImplementedError(nie)

    def metric(self, primals, tangents, **primals_kw):
        """Applies the metric at `primals` to `tangents`.

        Parameters
        ----------
        primals : tree-like structure
            Position at which to evaluate the metric.
        tangents : tree-like structure
            Instance to which to apply the metric.
        **primals_kw : Any
           Additional arguments passed on to the metric.

        Returns
        -------
        naturally_curved : tree-like structure
            Tree-like structure of the same type as primals to which the metric
            has been applied to.
        """
        lsm_at_p = Partial(self.left_sqrt_metric, primals, **primals_kw)
        return lsm_at_p(self.right_sqrt_metric(primals, tangents, **primals_kw))

    def left_sqrt_metric(self, primals, tangents, **primals_kw):
        """Applies the left-square-root (LSM) of the metric at `primals` to
        `tangents`.

        Parameters
        ----------
        primals : tree-like structure
            Position at which to evaluate the metric.
        tangents : tree-like structure
            Instance to which to apply the metric. Must be of shape
            `lsm_tangents_shape`.
        **primals_kw : Any
           Additional arguments passed on to the LSM.

        Returns
        -------
        metric_sample : tree-like structure
            Tree-like structure of the same type as primals to which
            the left-square-root of the metric has been applied to.
        """
        _, bwd = jax.vjp(Partial(self.transformation, **primals_kw), primals)
        bwd = _functional_conj(bwd)
        res = bwd(tangents)
        return res[0]

    def right_sqrt_metric(self, primals, tangents, **primals_kw):
        """Applies the right-square-root (RSM) of the metric at `primals` to
        `tangents`.

        Parameters
        ----------
        primals : tree-like structure
            Position at which to evaluate the metric.
        tangents : tree-like structure
            Instance to which to apply the metric. Must be of the same shape as
            `primals`.
        **primals_kw : Any
           Additional arguments passed on to the RSM.

        Returns
        -------
        metric_sample : tree-like structure
            Tree-like structure of the same type as
            `left_sqrt_metric_tangents_shape`.
        """
        lsm_at_p = Partial(self.left_sqrt_metric, primals, **primals_kw)
        rsm_at_p = jax.linear_transpose(lsm_at_p, self.left_sqrt_metric_tangents_shape)
        rsm_at_p = _functional_conj(rsm_at_p)
        return rsm_at_p(tangents)[0]

    def transformation(self, primals, **primals_kw):
        """Applies the coordinate transformation that maps into a coordinate
        system in which the metric of the likelihood is the Euclidean metric.

        Parameters
        ----------
        primals : tree-like structure
            Position at which to transform.
        **primals_kw : Any
           Additional arguments passed on to the transformation.

        Returns
        -------
        transformed_sample : tree-like structure
            Structure of the same type as lsm_tangents_shape to which the
            geometric transformation has been applied to.
        """
        raise NotImplementedError("`transformation` is not implemented")

    @property
    def left_sqrt_metric_tangents_shape(self):
        """Retrieves the shape of the tangent domain of the
        left-square-root of the metric.
        """
        return self._lsm_tan_shp

    @property
    def lsm_tangents_shape(self):
        """Alias for `left_sqrt_metric_tangents_shape`."""
        return self.left_sqrt_metric_tangents_shape

    @property
    def right_sqrt_metric_tangents_shape(self):
        """Retrieves the shape of the tangent domain of the
        right-square-root of the metric, a.k.a. the domain.
        """
        return self.domain

    @property
    def rsm_tangents_shape(self):
        """Alias for `right_sqrt_metric_tangents_shape`."""
        return self.right_sqrt_metric_tangents_shape

    def amend(self, f: Callable, /, *, domain=NoValue, likelihood_argnames=None):
        """Amend a forward model to the likelihood."""
        return LikelihoodWithModel(
            self, f, domain=domain, likelihood_argnames=likelihood_argnames
        )

    def __add__(self, other):
        return LikelihoodSum(self, other)

    def freeze(self, *, primals, point_estimates):
        """Returns a new likelihood with partially inserted `primals` and the
        remaining unfrozen/liquid `primals`.
        """
        if not point_estimates:
            return self, primals
        lp = LikelihoodPartial(self, primals=primals, point_estimates=point_estimates)
        return lp, lp.splitx(primals)[0]

    def __str__(self):
        return f"{self.__class__.__name__}()"


class LikelihoodPartial(Likelihood):
    """Likelihood with partially inserted `primals`."""

    likelihood: Likelihood = field(metadata=dict(static=False))
    primals_frozen: Any = field(metadata=dict(static=False))

    def __init__(
        self,
        likelihood,
        /,
        *,
        primals,
        point_estimates,
    ):
        self.likelihood = likelihood
        self.point_estimates = point_estimates
        self.insert_axes, pl, self.primals_frozen = _parse_point_estimates(
            self.point_estimates, primals
        )
        self.unflatten = Vector if self.insert_axes else None
        super().__init__(
            domain=tree_map(ShapeWithDtype.from_leave, pl),
            lsm_tangents_shape=self.likelihood.lsm_tangents_shape,
        )

    @property
    def energy(self):
        return partial_insert_and_remove(
            self.likelihood.energy,
            insert_axes=(self.insert_axes,),
            flat_fill=(self.primals_frozen,),
            remove_axes=None,
        )

    @property
    def transformation(self):
        return partial_insert_and_remove(
            self.likelihood.transformation,
            insert_axes=(self.insert_axes,),
            flat_fill=(self.primals_frozen,),
            remove_axes=None,
        )

    @property
    def left_sqrt_metric(self):
        return partial_insert_and_remove(
            self.likelihood.left_sqrt_metric,
            insert_axes=(self.insert_axes, None),
            flat_fill=(self.primals_frozen, None),
            remove_axes=self.insert_axes,
            unflatten=self.unflatten,
        )

    @property
    def right_sqrt_metric(self):
        return partial_insert_and_remove(
            self.likelihood.right_sqrt_metric,
            insert_axes=(self.insert_axes, self.insert_axes),
            flat_fill=(self.primals_frozen, zeros_like(self.primals_frozen)),
            remove_axes=None,
        )

    @property
    def metric(self):
        return partial_insert_and_remove(
            self.likelihood.metric,
            insert_axes=(self.insert_axes, self.insert_axes),
            flat_fill=(self.primals_frozen, zeros_like(self.primals_frozen)),
            remove_axes=self.insert_axes,
            unflatten=self.unflatten,
        )

    @property
    def normalized_residual(self):
        return partial_insert_and_remove(
            self.likelihood.normalized_residual,
            insert_axes=(self.insert_axes,),
            flat_fill=(self.primals_frozen,),
            remove_axes=None,
        )

    def splitx(self, primals):
        """Split the primals into liquid and frozen.

        Parameters
        ----------
        primals : tree-like structure
            Position at which to evaluate the energy.

        Returns
        -------
        primals_liqud : tree-like structure
            Liquid/dynamic part of the position at which to evaluate the energy.
        primals_frozen : tree-like structure
            Frozen/static part of the position at which to evaluate the energy.
        """
        return _parse_point_estimates(self.point_estimates, primals)[1:]

    def __str__(self):
        args_str = f"{self.likelihood}, point_estimates={self.point_estimates}"
        return f"{self.__class__.__name__}({args_str})"


class _ChainModel(LazyModel):
    forward_left: Callable = field(metadata=dict(static=False))
    forward_right: Callable = field(metadata=dict(static=False))

    def __init__(
        self,
        forward_left,
        forward_right,
        *,
        left_argnames=None,
        domain=NoValue,
        target=NoValue,
    ):
        self.forward_left = (
            forward_left
            if isinstance(forward_left, LazyModel)
            else Partial(forward_left)
        )
        self.forward_right = (
            forward_right
            if isinstance(forward_right, LazyModel)
            else Partial(forward_right)
        )
        left_argnames = () if left_argnames is None else left_argnames
        self._left_argnames = left_argnames

        domain = (
            forward_right.domain
            if domain is NoValue and isinstance(forward_right, LazyModel)
            else domain
        )
        target = (
            forward_left.target
            if target is NoValue and isinstance(forward_left, LazyModel)
            else target
        )
        super().__init__(domain=domain, target=target)

    def __call__(self, primals, **kwargs):
        kw_l = {k: kwargs.pop(k) for k in self._left_argnames}
        kw_r = kwargs
        return self.forward_left(self.forward_right(primals, **kw_r), **kw_l)


class LikelihoodWithModel(Likelihood):
    likelihood: Likelihood = field(metadata=dict(static=False))
    forward: Callable = field(metadata=dict(static=False))
    likelihood_argnames: tuple = ()

    def __init__(
        self,
        likelihood: Likelihood,
        f: Callable,
        /,
        *,
        domain=NoValue,
        init=NoValue,
        likelihood_argnames=None,
    ):
        """Amend the function `f` to the right of the likelihood.

        Parameters
        ----------
        f : Callable
            Function which to amend to the likelihood.
        likelihood_argnames : tuple or None
            Keys of the keyword arguments of the joined likelihood which
            to pass to the original likelihood. Passing `None` indicates
            the intent to absorb everything not explicitly absorbed by
            the other call.

        Returns
        -------
        lh : Likelihood
        """
        self.likelihood = likelihood
        if not callable(f):
            te = f"second argument to {self.__class__.__name__} must be callable; got {f!r}"
            raise TypeError(te)
        self.forward = f if isinstance(f, LazyModel) else Partial(f)
        likelihood_argnames = () if likelihood_argnames is None else likelihood_argnames
        if not isinstance(likelihood_argnames, (tuple, list)):
            te = f"invalid `likelihood_argnames` {self.likelihood_argnames!r}"
            raise TypeError(te)
        self.likelihood_argnames = likelihood_argnames
        domain = f.domain if domain is NoValue and isinstance(f, LazyModel) else domain
        init = f.init if init is NoValue and isinstance(f, LazyModel) else init
        super().__init__(
            domain=domain,
            init=init,
            lsm_tangents_shape=self.likelihood.lsm_tangents_shape,
        )

    def _split_kwargs(self, **kwargs):
        left = {k: kwargs.pop(k) for k in self.likelihood_argnames}
        return left, kwargs

    def energy(self, primals, **kwargs):
        kw_l, kw_r = self._split_kwargs(**kwargs)
        return self.likelihood(self.forward(primals, **kw_r), **kw_l)

    def normalized_residual(self, primals, **kwargs):
        kw_l, kw_r = self._split_kwargs(**kwargs)
        return self.likelihood.normalized_residual(
            self.forward(primals, **kw_r), **kw_l
        )

    def transformation(self, primals, **kwargs):
        kw_l, kw_r = self._split_kwargs(**kwargs)
        return self.likelihood.transformation(self.forward(primals, **kw_r), **kw_l)

    def metric(self, primals, tangents, **kwargs):
        kw_l, kw_r = self._split_kwargs(**kwargs)
        # Note, judging by a simple benchmark on a large problem,
        # transposing the JVP seems faster than computing the VJP again. On
        # small problems there seems to be no measurable difference.
        y, fwd = jax.linearize(Partial(self.forward, **kw_r), primals)
        bwd = jax.linear_transpose(fwd, primals)
        bwd = _functional_conj(bwd)
        return bwd(self.likelihood.metric(y, fwd(tangents), **kw_l))[0]

    def left_sqrt_metric(self, primals, tangents, **kwargs):
        kw_l, kw_r = self._split_kwargs(**kwargs)
        y, bwd = jax.vjp(Partial(self.forward, **kw_r), primals)
        bwd = _functional_conj(bwd)
        left_at_fp = self.likelihood.left_sqrt_metric(y, tangents, **kw_l)
        return bwd(left_at_fp)[0]

    def right_sqrt_metric(self, primals, tangents, **kwargs):
        kw_l, kw_r = self._split_kwargs(**kwargs)
        y, fwd = jax.linearize(Partial(self.forward, **kw_r), primals)
        return self.likelihood.right_sqrt_metric(y, fwd(tangents), **kw_l)

    def amend(
        self,
        f: Callable,
        *,
        domain=NoValue,
        left_argnames=None,
        likelihood_argnames=None,
    ):
        ff = _ChainModel(self.forward, f, left_argnames=left_argnames, domain=domain)
        likelihood_argnames = (
            self.likelihood_argnames
            if likelihood_argnames is None
            else likelihood_argnames
        )

        return self.__class__(
            self.likelihood,
            ff,
            domain=domain,
            likelihood_argnames=likelihood_argnames,
        )

    def __str__(self):
        return f"{self.likelihood}.amend({self.forward})"


class LikelihoodSum(Likelihood):
    likelihood_summands: Tuple[Likelihood] = field(metadata=dict(static=False))

    def __init__(
        self,
        *likelihood_summands,
        domain=NoValue,
        init=NoValue,
        _key_template="lh_{index}",
    ):
        for i, lh in enumerate(likelihood_summands):
            if not isinstance(lh, Likelihood):
                te = (
                    f"object at position {i} which to add to this instance is of"
                    f" invalid type {type(lh)!r}"
                )
                raise TypeError(te)
        self.likelihood_summands = tuple(likelihood_summands)
        self._key_template = _key_template

        joined_tangents_shape = {key: lh._lsm_tan_shp for key, lh in self._items()}
        if any(isinstance(lh._lsm_tan_shp, Vector) for _, lh in self._items()):
            joined_tangents_shape = Vector(joined_tangents_shape)

        if domain is NoValue and all(
            lh.domain is not NoValue for _, lh in self._items()
        ):
            domain = reduce(
                operator.or_,
                (
                    lh.domain.tree if isinstance(lh.domain, Vector) else lh.domain
                    for _, lh in self._items()
                ),
            )
            isvec = any(isinstance(lh.domain, Vector) for _, lh in self._items())
            domain = Vector(domain) if isvec else domain
            isswd = hasattr(domain, "shape") and hasattr(domain, "dtype")
            if not isswd and not has_arithmetics(domain):
                ve = (
                    "domains of the Likelihood-summands must support core"
                    " arithmetic operations"
                    "\nmaybe you forgot to wrap your inputs to the liklihoods"
                    " in `Vector`s"
                )
                raise ValueError(ve)
        super().__init__(
            domain=domain, init=init, lsm_tangents_shape=joined_tangents_shape
        )

    def _items(self):
        for i, lh in enumerate(self.likelihood_summands):
            # Allow the user to manipulate the keys to contain information about
            # the likelihood
            yield self._key_template.format(index=i, likelihood=lh), lh

    def energy(self, primals, **kwargs):
        return reduce(
            operator.add, (lh.energy(primals, **kwargs) for _, lh in self._items())
        )

    def normalized_residual(self, primals, **kwargs):
        res = {
            key: lh.normalized_residual(primals, **kwargs) for key, lh in self._items()
        }
        isvec = any(isinstance(lh.domain, Vector) for _, lh in self._items())
        return Vector(res) if isvec else res

    def metric(self, primals, tangents, **kwargs):
        return reduce(
            operator.add,
            (lh.metric(primals, tangents, **kwargs) for _, lh in self._items()),
        )

    def transformation(self, primals, **kwargs):
        res = {key: lh.transformation(primals, **kwargs) for key, lh in self._items()}
        isvec = any(isinstance(lh.domain, Vector) for _, lh in self._items())
        return Vector(res) if isvec else res

    def left_sqrt_metric(self, primals, tangents, **kwargs):
        return reduce(
            operator.add,
            (
                lh.left_sqrt_metric(primals, tangents[key], **kwargs)
                for key, lh in self._items()
            ),
        )

    def right_sqrt_metric(self, primals, tangents, **kwargs):
        res = {
            key: lh.right_sqrt_metric(primals, tangents, **kwargs)
            for key, lh in self._items()
        }
        isvec = any(isinstance(lh.domain, Vector) for _, lh in self._items())
        return Vector(res) if isvec else res

    def __add__(self, other):
        return LikelihoodSum(*self.likelihood_summands, other)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .. import utilities
from .diagonal_operator import DiagonalOperator
from .linear_operator import LinearOperator
from .scaling_operator import ScalingOperator
from .simple_linear_operators import NullOperator


class ChainOperator(LinearOperator):
    """Class representing chains of operators.

    Notes
    -----
    This operator has to be called using the :attr:`make` method.
    """

    def __init__(self, ops, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError
        self._ops = ops
        self._capability = self._all_ops
        for op in ops:
            self._capability &= op.capability
        self._domain = self._ops[-1].domain
        self._target = self._ops[0].target

    @staticmethod
    def simplify(ops):
        if len(ops) == 1:
            return ops
        # verify domains
        for i in range(len(ops) - 1):
            utilities.check_object_identity(ops[i + 1].target, ops[i].domain)
        if len(ops)==2:
            if ops[0].isIdentity():
                return [ops[1]]
            if ops[1].isIdentity():
                return [ops[0]]
        # unpack ChainOperators
        opsnew = []
        for op in ops:
            opsnew += op._ops if isinstance(op, ChainOperator) else [op]
        ops = opsnew
        # check for NullOperators
        if any(isinstance(op, NullOperator) for op in ops):
            ops = (NullOperator(ops[-1].domain, ops[0].target),)
        # collect ScalingOperators
        fct = 1.
        opsnew = []
        lastdom = ops[-1].domain
        dtype = None
        for op in ops:
            if (isinstance(op, ScalingOperator) and op._factor.imag == 0):
                fct *= op._factor.real
            else:
                opsnew.append(op)
        if fct != 1.:
            # try to absorb the factor into a DiagonalOperator
            for i in range(len(opsnew)):
                if isinstance(opsnew[i], DiagonalOperator):
                    opsnew[i] = opsnew[i]._scale(fct)
                    fct = 1.
                    break
        if fct != 1 or len(opsnew) == 0:
            # have to add the scaling operator at the end
            op = ScalingOperator(lastdom, fct)
            if dtype is not None:
                op.dtype = dtype
            opsnew.append(op)
        ops = opsnew
        # combine DiagonalOperators where possible
        opsnew = []
        for op in ops:
            if (len(opsnew) > 0 and isinstance(opsnew[-1], DiagonalOperator)
                    and isinstance(op, DiagonalOperator)):
                opsnew[-1] = opsnew[-1]._combine_prod(op)
            else:
                opsnew.append(op)
        ops = opsnew
        # combine BlockDiagonalOperators where possible
        from .block_diagonal_operator import BlockDiagonalOperator
        opsnew = []
        for op in ops:
            if (len(opsnew) > 0
                    and isinstance(opsnew[-1], BlockDiagonalOperator)
                    and isinstance(op, BlockDiagonalOperator)):
                opsnew[-1] = opsnew[-1]._combine_chain(op)
            else:
                opsnew.append(op)
        ops = opsnew
        return ops

    @staticmethod
    def make(ops):
        """Build a ChainOperator (or something simpler if possible),
        a sequence of concatenated LinearOperators.

        Parameters
        ----------
        ops: list of LinearOperator
            Individual operators of the chain.
        """
        ops = tuple(ops)
        if len(ops) == 0:
            raise ValueError("ops is empty")
        ops = ChainOperator.simplify(ops)
        if len(ops) == 1:
            return ops[0]
        return ChainOperator(ops, _callingfrommake=True)

    def _flip_modes(self, trafo):
        ADJ = self.ADJOINT_BIT
        INV = self.INVERSE_BIT

        if trafo == 0:
            return self
        if trafo == ADJ or trafo == INV:
            return self.make(
                [op._flip_modes(trafo) for op in reversed(self._ops)])
        if trafo == ADJ | INV:
            return self.make([op._flip_modes(trafo) for op in self._ops])
        raise ValueError("invalid operator transformation")

    def apply(self, x, mode):
        self._check_mode(mode)
        t_ops = self._ops if mode & self._backwards else reversed(self._ops)
        for op in t_ops:
            x = op.apply(x, mode)
        return x

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in self._ops)
        return "ChainOperator:\n" + utilities.indent(subs)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from ..multi_domain import MultiDomain
        if not isinstance(self._domain, MultiDomain):
            return None, self
        newop = None
        for op in reversed(self._ops):
            c_inp, t_op = op.simplify_for_constant_input(c_inp)
            newop = t_op if newop is None else op(newop)
        return c_inp, newop



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..field import Field
from .linear_operator import LinearOperator


class DomainTupleFieldInserter(LinearOperator):
    """Writes the content of a :class:`Field` into one slice of a
    :class:`DomainTuple`.

    Parameters
    ----------
    target : Domain, tuple of Domain or DomainTuple
    space : int
       The index of the sub-domain which is inserted.
    index : tuple
        Slice in new sub-domain in which the input field shall be written into.
    """

    def __init__(self, target, space, index):
        if not space <= len(target) or space < 0:
            raise ValueError
        self._target = DomainTuple.make(target)
        dom = list(self.target)
        dom.pop(space)
        self._domain = DomainTuple.make(dom)
        self._capability = self.TIMES | self.ADJOINT_TIMES

        new_space = target[space]
        nshp = new_space.shape
        fst_dims = sum(len(dd.shape) for dd in self.target[:space])

        if len(index) != len(nshp):
            raise ValueError("shape mismatch between new_space and position")
        for s, p in zip(nshp, index):
            if p < 0 or p >= s:
                raise ValueError("bad position value")

        self._slc = (slice(None),)*fst_dims + index

    def apply(self, x, mode):
        self._check_input(x, mode)
        x = x.val
        if mode == self.TIMES:
            res = np.zeros_like(x, shape=self.target.shape, dtype=x.dtype)
            res[self._slc] = x
            return Field(self.target, res)
        else:
            return Field(self.domain, x[self._slc])



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import operator

from jax.tree_util import register_pytree_node_class, tree_map


def _unary_op(op, name=None):
    def unary_call(lhs):
        return op(lhs._str)

    name = op.__name__ if name is None else name
    unary_call.__name__ = f"__{name}__"
    return unary_call


def _binary_op(op, name=None):
    def binary_call(lhs, rhs):
        lhs = lhs._str if isinstance(lhs, PyTreeString) else lhs
        rhs = rhs._str if isinstance(rhs, PyTreeString) else rhs
        out = op(lhs, rhs)
        return PyTreeString(out) if isinstance(out, str) else out

    name = op.__name__ if name is None else name
    binary_call.__name__ = f"__{name}__"
    return binary_call


def _rev_binary_op(op, name=None):
    def binary_call(lhs, rhs):
        lhs = lhs._str if isinstance(lhs, PyTreeString) else lhs
        rhs = rhs._str if isinstance(rhs, PyTreeString) else rhs
        out = op(rhs, lhs)
        return PyTreeString(out) if isinstance(out, str) else out

    name = op.__name__ if name is None else name
    binary_call.__name__ = f"__r{name}__"
    return binary_call


def _fwd_rev_binary_op(op, name=None):
    return (_binary_op(op, name=name), _rev_binary_op(op, name=name))


@register_pytree_node_class
class PyTreeString:
    def __init__(self, str):
        self._str = str

    def tree_flatten(self):
        return ((), (self._str,))

    @classmethod
    def tree_unflatten(cls, aux, _):
        return cls(*aux)

    def __str__(self):
        return self._str

    def __repr__(self):
        return f"{self.__class__.__name__}({self._str!r})"

    __lt__ = _binary_op(operator.lt)
    __le__ = _binary_op(operator.le)
    __eq__ = _binary_op(operator.eq)
    __ne__ = _binary_op(operator.ne)
    __ge__ = _binary_op(operator.ge)
    __gt__ = _binary_op(operator.gt)

    __add__, __radd__ = _fwd_rev_binary_op(operator.add)
    __mul__, __rmul__ = _fwd_rev_binary_op(operator.mul)

    lower = _unary_op(str.lower)
    upper = _unary_op(str.upper)

    __hash__ = _unary_op(str.__hash__)

    startswith = _binary_op(str.startswith)


def hide_strings(a):
    return tree_map(lambda x: PyTreeString(x) if isinstance(x, str) else x, a)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2025 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .. import utilities
from ..any_array import AnyArray
from ..domain_tuple import DomainTuple
from ..field import Field
from .endomorphic_operator import EndomorphicOperator

# TODO: Eventually enforce somewhat modern ducc version (>=0.37.0) as nifty
# dependency and remove the try statement below (also at other places)
try:
    from ducc0.misc.experimental import div_conj, mul_conj
    def mul_conj2(a, b):
        assert a.device_id == b.device_id
        if a.device_id == -1:
            return AnyArray(mul_conj(a.val, b.val))
        return a*b.conj()

    def div_conj2(a, b):
        assert a.device_id == b.device_id
        if a.device_id == -1:
            return AnyArray(div_conj(a.val, b.val))
        return a/b.conj()

except ImportError:
    def mul_conj2(a, b):
        return a*b.conj()

    def div_conj2(a, b):
        return a/b.conj()


class DiagonalOperator(EndomorphicOperator):
    """Represents a :class:`LinearOperator` which is diagonal.

    The NIFTy DiagonalOperator class is a subclass derived from the
    :class:`EndomorphicOperator`. It multiplies an input field pixel-wise with
    its diagonal.

    Parameters
    ----------
    diagonal : :class:`nifty.cl.field.Field`
        The diagonal entries of the operator.
    domain : Domain, tuple of Domain or DomainTuple, optional
        The domain on which the Operator's input Field is defined.
        If None, use the domain of "diagonal".
    spaces : int or tuple of int, optional
        The elements of "domain" on which the operator acts.
        If None, it acts on all elements.
    sampling_dtype :
        If this operator represents the covariance of a Gaussian probabilty
        distribution, `sampling_dtype` specifies if it is real or complex
        Gaussian. If `sampling_dtype` is `None`, the operator cannot be used as
        a covariance, i.e. no samples can be drawn. Default: None.

    Notes
    -----
    Formally, this operator always supports all operation modes (times,
    adjoint_times, inverse_times and inverse_adjoint_times), even if there
    are diagonal elements with value 0 or infinity. It is the user's
    responsibility to apply the operator only in appropriate ways (e.g. call
    inverse_times only if there are no zeros on the diagonal).

    This shortcoming will hopefully be fixed in the future.
    """

    def __init__(self, diagonal, domain=None, spaces=None, sampling_dtype=None,
                 _trafo=0):
# MR: _trafo is more or less deliberately undocumented, since it is not supposed
# to be necessary for "end users". It describes the type of transform for which
# the diagonal can be used without modification
# (0:TIMES, 1:ADJOINT, 2:INVERSE, 3:ADJOINT_INVERSE)
        if not isinstance(diagonal, Field):
            raise TypeError("Field object required")
        utilities.check_dtype_or_none(sampling_dtype)
        self._dtype = sampling_dtype
        self._trafo = _trafo
        if domain is None:
            self._domain = diagonal.domain
        else:
            self._domain = DomainTuple.make(domain)
        if spaces is None:
            self._spaces = None
            utilities.check_object_identity(diagonal.domain, self._domain)
        else:
            self._spaces = utilities.parse_spaces(spaces, len(self._domain))
            if len(self._spaces) != len(diagonal.domain):
                raise ValueError("spaces and domain must have the same length")
            for i, j in enumerate(self._spaces):
                if diagonal.domain[i] != self._domain[j]:
                    raise ValueError(f"Mismatch between:\n{diagonal.domain[i]}\n"
                                     f"and:\n{self._domain[j]}")
            if self._spaces == tuple(range(len(self._domain))):
                self._spaces = None  # shortcut

        if self._spaces is not None:
            active_axes = []
            for space_index in self._spaces:
                active_axes += self._domain.axes[space_index]

            self._ldiag = diagonal.val
            self._reshaper = [shp if i in active_axes else 1
                              for i, shp in enumerate(self._domain.shape)]
            self._ldiag = self._ldiag.reshape(self._reshaper)
        else:
            self._ldiag = diagonal.val
        assert isinstance(self._ldiag, AnyArray)
        self._fill_rest()

    def _fill_rest(self):
        self._ldiag.lock()
        self._complex = utilities.iscomplextype(self._ldiag.dtype)
        self._capability = self._all_ops
        if not self._complex:
            self._diagmin_cache = None

    @property
    def _diagmin(self):
        if self._complex:
            raise RuntimeError("complex DiagonalOperator does not have _diagmin")
        if self._diagmin_cache is None:
            self._diagmin_cache = self._ldiag.min()
        return self._diagmin_cache

    def _from_ldiag(self, spc, ldiag, sampling_dtype, trafo):
        res = DiagonalOperator.__new__(DiagonalOperator)
        res._dtype = sampling_dtype
        res._trafo = trafo
        res._domain = self._domain
        if self._spaces is None or spc is None:
            res._spaces = None
        else:
            res._spaces = tuple(set(self._spaces) | set(spc))
        res._ldiag = AnyArray(ldiag)
        res._fill_rest()
        return res

    def _get_actual_diag(self):
        if self._trafo == 0:
            return self._ldiag
        if self._trafo == 1:
            return np.conj(self._ldiag) if self._complex else self._ldiag
        if self._trafo == 2:
            return 1./self._ldiag
        if self._trafo == 3:
            return np.conj(1./self._ldiag) if self._complex else 1./self._ldiag

    def _scale(self, fct):
        if not np.isscalar(fct):
            raise TypeError("scalar value required")
        return self._from_ldiag((), self._get_actual_diag()*fct, self._dtype, 0)

    def _add(self, sum_):
        if not np.isscalar(sum_):
            raise TypeError("scalar value required")
        return self._from_ldiag((), self._get_actual_diag()+sum_, self._dtype, 0)

    def _combine_prod(self, op):
        if not isinstance(op, DiagonalOperator):
            raise TypeError("DiagonalOperator required")
        dtype = self._dtype if self._dtype == op._dtype else None
        return self._from_ldiag(op._spaces, self._get_actual_diag()*op._get_actual_diag(),
                                dtype, 0)

    def _combine_sum(self, op, selfneg, opneg):
        if not isinstance(op, DiagonalOperator):
            raise TypeError("DiagonalOperator required")
        tdiag = (self._get_actual_diag() * (-1 if selfneg else 1) +
                 op._get_actual_diag() * (-1 if opneg else 1))
        dtype = self._dtype if self._dtype == op._dtype else None
        return self._from_ldiag(op._spaces, tdiag, dtype, 0)

    def _device_preparation(self, x, mode):
        self._ldiag = self._ldiag.at(x.device_id)

    def apply(self, x, mode):
        self._check_input(x, mode)
        # To save both time and memory, we remap the `mode` (via `self._trafo`)
        # and do not compute and store a new `self._ldiag`s for adjoint, inverse
        # or adjoint-inverse DiagonalOperators.
        self._device_preparation(x, mode)
        trafo = self._ilog[mode] ^ self._trafo

        if trafo == 0:  # straight application
            return Field(x.domain, x.val*self._ldiag)

        if trafo == 1:  # adjoint
            return Field(x.domain, mul_conj2(x.val, self._ldiag)
                                   if self._complex else x.val*self._ldiag)

        if trafo == 2:  # inverse
            return Field(x.domain, x.val/self._ldiag)

        # adjoint inverse
        return Field(x.domain, div_conj2(x.val, self._ldiag)
            if self._complex else x.val/self._ldiag)

    def _flip_modes(self, trafo):
        return self._from_ldiag((), self._ldiag, self._dtype, self._trafo ^ trafo)

    def process_sample(self, samp, from_inverse):
        from_inverse2 = from_inverse ^ (self._trafo >= 2)
        # `from_inverse2` captures if the inverse of `self._ldiag` needs to be
        # taken or not (can happen for nontrivial `self._trafo`).
        if (self._complex or (self._diagmin < 0.) or
                (self._diagmin == 0. and from_inverse2)):
            raise ValueError("operator not positive definite")
        if from_inverse2:
            res = samp.val/np.sqrt(self._ldiag)
        else:
            res = samp.val*np.sqrt(self._ldiag)
        return Field(self._domain, res)

    def draw_sample(self, from_inverse=False, device_id=-1):
        if self._dtype is None:
            s = "Need to specify dtype to be able to sample from this operator:\n"
            s += self.__repr__()
            raise RuntimeError(s)
        res = Field.from_random(domain=self._domain, random_type="normal",
                                dtype=self._dtype, device_id=device_id)
        return self.process_sample(res, from_inverse)

    def get_sqrt(self):
        if self._complex or self._diagmin < 0.:
            raise ValueError("get_sqrt() works only for positive definite operators.")
        return self._from_ldiag((), np.sqrt(self._ldiag), self._dtype, self._trafo)

    def __repr__(self):
        from ..multi_domain import MultiDomain
        s = "DiagonalOperator (domain/target "
        if isinstance(self.domain, MultiDomain):
            s += f"keys: {self._domain.keys()}"
        else:
            s += f"shape: {self._domain.shape}"
        if self._dtype is not None:
            s += f", sampling dtype {self._dtype}"
        s += ")"
        return s



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..field import Field
from ..operators.linear_operator import LinearOperator
from ..operators.operator import Operator


def _field_from_function(domain, func, absolute=False):
    domain = DomainTuple.make(domain)
    k_array = _make_coords(domain, absolute=absolute)
    return Field(domain, func(k_array))


def _make_coords(domain, absolute=False):
    domain = DomainTuple.make(domain)
    dim = len(domain.shape)
    dist = domain[0].distances
    shape = domain.shape
    k_array = np.zeros((dim,) + shape)
    for i in range(dim):
        ks = np.minimum(shape[i] - np.arange(shape[i]), np.arange(
            shape[i]))*dist[i]
        if not absolute:
            ks[int(shape[i]/2) + 1:] *= -1
        fst_dims = (1,)*i
        lst_dims = (1,)*(dim - i - 1)
        k_array[i] += ks.reshape(fst_dims + (shape[i],) + lst_dims)
    return k_array


class _LightConeDerivative(LinearOperator):
    def __init__(self, domain, target, derivatives):
        super(_LightConeDerivative, self).__init__()
        self._domain = domain
        self._target = target
        self._derivatives = derivatives
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def _device_preparation(self, x, mode):
        self._derivatives = self._derivatives.at(x.device_id)

    def apply(self, x, mode):
        self._check_input(x, mode)
        self._device_preparation(x, mode)
        x = x.val
        res = np.zeros_like(x, shape=self._tgt(mode).shape, dtype=self._derivatives.dtype)
        for i in range(self.domain.shape[0]):
            if mode == self.TIMES:
                res += self._derivatives[i]*x[i]
            else:
                res[i] = np.sum(self._derivatives[i]*x.real)
        return Field(self._tgt(mode), res)


def _cone_arrays(c, domain, sigx, want_gradient):
    x = _make_coords(domain)
    from ..any_array import AnyArray
    x = AnyArray(x).at(c.device_id)  # TEMP
    a = np.zeros_like(c, shape=domain.shape, dtype=np.complex128)
    if want_gradient:
        derivs = np.zeros_like(c, shape=(c.size,) + domain.shape, dtype=np.complex128)
    else:
        derivs = None
    a -= (x[0]/(sigx*domain[0].distances[0]))**2
    for i in range(c.size):
        res = (x[i + 1]/(sigx*domain[0].distances[i + 1]))**2
        a += c[i]*res
        if want_gradient:
            derivs[i] = res
    a = np.sqrt(a)
    if want_gradient:
        derivs *= -0.5
        for i in range(c.size):
            derivs[i][a == 0] = 0.
            derivs[i][a != 0] /= a[a != 0]
    a = a.real
    if want_gradient:
        derivs *= a
    a = np.exp(-0.5*a**2)
    if want_gradient:
        derivs = a*derivs.real
    return a, derivs


class LightConeOperator(Operator):
    '''Constructs a Light cone from a set of lightspeed parameters.

    The resulting cone is defined as follows

    .. math::
        \\exp \\left(- \\frac{1}{2} \\Re \\left( \\Delta \\right)^2 \\right)

    with

    .. math::
        \\Delta = \\sqrt{- \\left(t^2 - \\frac{x^\\dagger C^{-1} x}
        {\\sigma_x^2} \\right)}

    where t and x are the coordinates of the target space. Note that axis zero
    of the space is interpreted as the time axis. C denotes the input
    paramters of the operator and parametrizes the shape of the cone.
    sigx is the width of the asymptotic Gaussian in x necessary for
    discretization.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain of the input parameters of the light cone, the values of the
        lightspeed tensor.
    target : Domain, tuple of Domain or DomainTuple
        Output space on which the lightcone should be defined. The zeroth axis
        of this space is interpreted as the time axis.
    sigx : float
        Width of the Gaussian for the discretized representation of the cone.
    '''
    def __init__(self, domain, target, sigx):
        self._domain = DomainTuple.make(domain)
        self._target = DomainTuple.make(target)
        self._sigx = sigx

    def apply(self, x):
        lin = x.jac is not None
        a, derivs = _cone_arrays(x.val.val if lin else x.val, self.target, self._sigx, lin)
        res = Field(self.target, a)
        if not lin:
            return res
        jac = _LightConeDerivative(self._domain, self._target, derivs)
        return x.new(res, jac)



# Copyright(C) 2023 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Authors: Gordian Edenhofer, Philipp Frank

from functools import partial
from operator import getitem
from typing import Callable, Optional, Tuple, TypeVar, Union

import jax
from jax import numpy as jnp
from jax import random
from jax.tree_util import Partial, register_pytree_node_class, tree_leaves, tree_map

from . import conjugate_gradient, optimize
from .likelihood import (
    Likelihood,
    LikelihoodWithModel,
    _functional_conj,
    _parse_point_estimates,
    partial_insert_and_remove,
)
from .misc import conditional_raise
from .tree_math import (
    Vector,
    assert_arithmetics,
    get_map,
    random_like,
    stack,
    vdot,
    conj,
    zeros_like,
)

P = TypeVar("P")


def _no_jit(x, **kwargs):
    return x


def _parse_jit(jit):
    if callable(jit):
        return jit
    if isinstance(jit, bool):
        return jax.jit if jit else _no_jit
    raise TypeError(f"expected `jit` to be callable or boolean; got {jit!r}")


@jax.jit
def concatenate_zip(*arrays):
    return tree_map(
        lambda *x: jnp.stack(x, axis=1).reshape((-1,) + x[0].shape[1:]), *arrays
    )


def _process_point_estimate(x, primals, point_estimates, insert):
    if not point_estimates:
        return x

    point_estimates, _, p_frozen = _parse_point_estimates(point_estimates, primals)
    assert p_frozen is not None
    fill = tree_map(lambda x: jnp.zeros((1,) * jnp.ndim(x)), p_frozen)
    in_out = partial_insert_and_remove(
        lambda *x: x[0],
        insert_axes=(point_estimates,) if insert else None,
        flat_fill=(fill,) if insert else None,
        remove_axes=None if insert else (point_estimates,),
        unflatten=None if insert else Vector,
    )
    return in_out(x)


def sample_likelihood(likelihood: Likelihood, primals, key):
    white_sample = random_like(key, likelihood.left_sqrt_metric_tangents_shape)
    return likelihood.left_sqrt_metric(primals, white_sample)


def draw_linear_residual(
    likelihood: Likelihood,
    pos: P,
    key,
    *,
    from_inverse: bool = True,
    point_estimates: Union[P, Tuple[str]] = (),
    cg: Callable = conjugate_gradient.static_cg,
    cg_name: Optional[str] = None,
    cg_kwargs: Optional[dict] = None,
    _raise_nonposdef: bool = False,
) -> tuple[P, int]:
    assert_arithmetics(pos)

    if not isinstance(likelihood, Likelihood):
        te = f"`likelihood` of invalid type; got '{type(likelihood)}'"
        raise TypeError(te)
    lh, p_liquid = likelihood, pos
    if point_estimates:
        lh, p_liquid = likelihood.freeze(point_estimates=point_estimates, primals=pos)

    def ham_metric(primals, tangents, **primals_kw):
        return lh.metric(primals, tangents, **primals_kw) + tangents

    cg_kwargs = cg_kwargs if cg_kwargs is not None else {}

    subkey_nll, subkey_prr = random.split(key, 2)
    nll_smpl = sample_likelihood(lh, p_liquid, key=subkey_nll)
    prr_inv_metric_smpl = random_like(key=subkey_prr, primals=p_liquid)
    # One may transform any metric sample to a sample of the inverse
    # metric by simply applying the inverse metric to it
    prr_smpl = prr_inv_metric_smpl
    # Note, we can sample antithetically by swapping the global sign of
    # the metric sample below (which corresponds to mirroring the final
    # sample) and additionally by swapping the relative sign between
    # the prior and the likelihood sample. The first technique is
    # computationally cheap and empirically known to improve stability.
    # The latter technique requires an additional inversion and its
    # impact on stability is still unknown.
    # TODO: investigate the impact of sampling the prior and likelihood
    # antithetically.
    smpl = nll_smpl + prr_smpl
    info = 0
    if from_inverse:
        inv_metric_at_p = partial(
            cg,
            Partial(ham_metric, p_liquid),
            **{"name": cg_name, "_raise_nonposdef": _raise_nonposdef, **cg_kwargs},
        )
        smpl, info = inv_metric_at_p(smpl, x0=prr_inv_metric_smpl)
        conditional_raise(
            (info < 0) if info is not None else False,
            ValueError("conjugate gradient failed"),
        )
    smpl = _process_point_estimate(smpl, pos, point_estimates, insert=True)
    return smpl, info


def nonlinearly_update_residual(
    likelihood=None,
    pos: P = None,
    residual_sample=None,
    metric_sample_key=None,
    metric_sample_sign=1.0,
    *,
    point_estimates=(),
    minimize: Callable[..., optimize.OptimizeResults] = optimize._static_newton_cg,
    minimize_kwargs={},
    _raise_notconverged=False,
) -> tuple[P, optimize.OptimizeResults]:
    assert_arithmetics(pos)
    assert_arithmetics(residual_sample)

    draw_lni = partial(
        draw_linear_residual,
        likelihood,
        from_inverse=False,
        point_estimates=point_estimates,
    )

    def residual_vg(e, lh_trafo_at_p, ms_at_p, x):
        lh, e_liquid = likelihood.freeze(point_estimates=point_estimates, primals=e)

        # t = likelihood.transformation(x) - lh_trafo_at_p
        t = tree_map(jnp.subtract, lh.transformation(x), lh_trafo_at_p)
        g = x - e_liquid + lh.left_sqrt_metric(e_liquid, t)
        r = ms_at_p - g
        res = 0.5 * vdot(r, r)

        r = conj(r)
        ngrad = r + lh.left_sqrt_metric(x, lh.right_sqrt_metric(e_liquid, r))
        return (res, -ngrad)

    def metric(e, primals, tangents):
        lh, e_liquid = likelihood.freeze(point_estimates=point_estimates, primals=e)
        lsm = lh.left_sqrt_metric
        rsm = lh.right_sqrt_metric
        tm = lsm(e_liquid, rsm(primals, tangents)) + tangents
        return lsm(primals, rsm(e_liquid, tm)) + tm

    def sampnorm(e, natgrad):
        lh, e_liquid = likelihood.freeze(point_estimates=point_estimates, primals=e)
        fpp = lh.right_sqrt_metric(e_liquid, natgrad)
        return jnp.sqrt(vdot(natgrad, natgrad) + vdot(fpp, fpp))

    sample = pos + residual_sample
    del residual_sample
    sample = _process_point_estimate(sample, pos, point_estimates, insert=False)
    metric_sample, _ = draw_lni(pos, metric_sample_key)
    metric_sample *= metric_sample_sign
    metric_sample = _process_point_estimate(
        metric_sample, pos, point_estimates, insert=False
    )
    # HACK for skipping the nonlinear update steps and not calling trafo
    skip = (
        isinstance(minimize_kwargs.get("maxiter", None), int)
        and minimize_kwargs["maxiter"] == 0
    )
    if not skip:
        lh_f, e_liquid = likelihood.freeze(point_estimates=point_estimates, primals=pos)
        trafo_at_p = lh_f.transformation(e_liquid)
        options = {
            "fun_and_grad": partial(residual_vg, pos, trafo_at_p, metric_sample),
            "hessp": partial(metric, pos),
            "custom_gradnorm": partial(sampnorm, pos),
        }
        opt_state = minimize(None, x0=sample, **(minimize_kwargs | options))
    else:
        opt_state = optimize.OptimizeResults(sample, True, 0, None, None)
    if _raise_notconverged and (opt_state.status < 0):
        ValueError("S: failed to invert map")
    # Subtract position in the reduced space (i.e. space w/o point-estimates) to
    # not pollute the point-estimated parameters with the mean
    sample = opt_state.x - _process_point_estimate(
        pos, pos, point_estimates, insert=False
    )
    # Remove x from state to avoid copy of the samples
    opt_state = opt_state._replace(x=None, jac=None)
    sample = _process_point_estimate(sample, pos, point_estimates, insert=True)
    return sample, opt_state


def draw_residual(
    likelihood: Likelihood,
    pos: P,
    key,
    *,
    point_estimates: Union[P, Tuple[str]] = (),
    cg: Callable = conjugate_gradient.static_cg,
    cg_name: Optional[str] = None,
    cg_kwargs: Optional[dict] = None,
    minimize: Callable[..., optimize.OptimizeResults] = optimize._static_newton_cg,
    minimize_kwargs={},
    _raise_nonposdef: bool = False,
    _raise_notconverged: bool = False,
) -> tuple[P, optimize.OptimizeResults]:
    residual_sample, _ = draw_linear_residual(
        likelihood,
        pos,
        key,
        point_estimates=point_estimates,
        cg=cg,
        cg_name=cg_name,
        cg_kwargs=cg_kwargs,
        _raise_nonposdef=_raise_nonposdef,
    )
    curve = partial(
        nonlinearly_update_residual,
        likelihood,
        pos,
        metric_sample_key=key,
        point_estimates=point_estimates,
        minimize=minimize,
        minimize_kwargs=minimize_kwargs,
        _raise_notconverged=_raise_notconverged,
    )
    return stack(
        (
            curve(residual_sample, metric_sample_sign=1.0),
            curve(-residual_sample, metric_sample_sign=-1.0),
        )
    )


@register_pytree_node_class
class Samples:
    """Storage class for samples (relative to some expansion point) that is
    fully compatible with JAX transformations like vmap, pmap, etc.

    This class is used to store samples for the Variational Inference schemes
    MGVI and geoVI where samples are defined relative to some expansion point
    (a.k.a. latent mean or offset).

    See also
    --------
    `Geometric Variational Inference`, Philipp Frank, Reimar Leike,
    Torsten A. Enßlin, `<https://arxiv.org/abs/2105.10470>`_
    `<https://doi.org/10.3390/e23070853>`_

    `Metric Gaussian Variational Inference`, Jakob Knollmüller,
    Torsten A. Enßlin, `<https://arxiv.org/abs/1901.11033>`_
    """

    def __init__(self, *, pos: P = None, samples: P, keys=None):
        self._pos, self._samples, self._keys = pos, samples, keys

    @property
    def pos(self):
        return self._pos

    @property
    def samples(self):
        if self._samples is None:
            raise ValueError(f"{self.__class__.__name__} has no samples")

        smpls = self._samples
        if self.pos is not None:
            smpls = tree_map(lambda p, s: p[jnp.newaxis] + s, self.pos, smpls)
        return smpls

    @property
    def keys(self):
        return self._keys

    def __len__(self):
        if self._samples is None:
            return 0
        return jnp.shape(tree_leaves(self._samples)[0])[0]

    def __getitem__(self, index):
        if self._samples is None:
            raise ValueError(f"{self.__class__.__name__} has no samples")

        def get(b):
            return getitem(b, index)

        if self.pos is None:
            return tree_map(get, self._samples)
        return tree_map(lambda p, s: p + get(s), self.pos, self._samples)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.samples == other.samples

    def at(self, pos, old_pos=None):
        """Update the offset (usually the latent mean) of all samples and
        optionally subtracts `old_pos` from all samples before.
        """
        if self.pos is not None and old_pos is None:
            smpls = self._samples
        elif old_pos is not None:
            smpls = self.samples
            smpls = tree_map(lambda p, s: s - p[jnp.newaxis], old_pos, smpls)
        else:
            raise ValueError("invalid combination of `pos` and `old_pos`")
        return Samples(pos=pos, samples=smpls, keys=self.keys)

    def squeeze(self):
        """Convenience method to merge the two leading axis of stacked samples
        (e.g. from batching).
        """
        smpls = tree_map(lambda s: s.reshape((-1,) + s.shape[2:]), self._samples)
        return Samples(pos=self.pos, samples=smpls, keys=self.keys)

    def tree_flatten(self):
        # Include mean in samples when passing to JAX (for e.g. vmap, pmap, ...)
        # return ((self.samples, ), (self.pos, ))  # confuses JAX
        return ((self.pos, self._samples, self.keys), ())

    @classmethod
    def tree_unflatten(cls, aux, children):
        # pos, = aux
        pos, smpls, keys = children
        # if pos is not None:  # confuses JAX
        #     smpls = tree_map(lambda p, s: s - p[jnp.newaxis], pos, smpls)
        return cls(pos=pos, samples=smpls, keys=keys)


def wiener_filter_posterior(
    likelihood: LikelihoodWithModel,
    position: Optional[P] = None,
    *,
    key,
    n_samples: int = 0,
    residual_map="smap",
    draw_linear_kwargs: Optional[dict] = None,
    optimize_for_linear: bool = False,
) -> Tuple[Samples, Tuple]:
    """Computes wiener filter solution for a standardized model.

    Parameters
    ----------
    likelihood : :class:`~nifty.re.likelihood.LikelihoodWithModel`
        Likelihood to be used for the wiener filter.
    primals : tree-like
        Position around which to linearize (if the problem is non-linear).
    key : jax random number generation key
    n_samples : int
        Number of samples to draw.
    residual_map: callable
        Map function used for the residual sample drawing.
    draw_linear_kwargs : dict
        Optional parameters passed on to :func:`draw_linear_residual` if these
        should not be the same as for the retrieval of the posterior mean.
    optimize_for_linear: bool
        Whether to optimize computations for linear model.
    """
    if not isinstance(likelihood, LikelihoodWithModel):
        msg = f"likelihood must be of LikelihoodWithModel type; got {likelihood}"
        return TypeError(msg)
    residual_map = get_map(residual_map)

    data = likelihood.likelihood.data
    # Remove any constant offsets from the data/signal that are part of the model
    data = data - likelihood.forward(zeros_like(likelihood.domain))

    position = zeros_like(likelihood.domain) if position is None else position
    if optimize_for_linear:
        forward_T = jax.linear_transpose(likelihood.forward, likelihood.domain)
    else:
        _, forward_T = jax.vjp(likelihood.forward, position)
    forward_T = _functional_conj(forward_T)
    n_inv_d = likelihood.likelihood.metric(likelihood.forward(position), data)
    (j,) = forward_T(n_inv_d)

    def post_cov_inv(tangents, primals):
        return likelihood.metric(primals, tangents) + tangents

    cg = draw_linear_kwargs.get("cg", conjugate_gradient.static_cg)
    post_mean, post_info = cg(
        Partial(post_cov_inv, primals=position),
        j,
        name=draw_linear_kwargs.get("cg_name", None),
        **draw_linear_kwargs.get("cg_kwargs", {}),
    )
    if post_info is not None and post_info < 0:
        raise ValueError("conjugate gradient failed")

    ks = random.split(key, n_samples)
    draw = Partial(draw_linear_residual, likelihood, **draw_linear_kwargs)
    draw = residual_map(draw, in_axes=(None, 0))
    smpls, smpls_info = draw(post_mean, ks)

    smpls = Samples(pos=post_mean, samples=concatenate_zip(smpls, -smpls), keys=ks)
    return smpls, (post_info, smpls_info)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from collections import defaultdict

from ..sugar import domain_union
from ..utilities import indent
from .block_diagonal_operator import BlockDiagonalOperator
from .linear_operator import LinearOperator


class SumOperator(LinearOperator):
    """Class representing sums of operators.

    Notes
    -----
    This operator has to be called using the `make` method.
    """

    def __init__(self, ops, neg, dom, tgt, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError
        self._ops = ops
        self._neg = neg
        self._domain = dom
        self._target = tgt
        self._capability = self.TIMES | self.ADJOINT_TIMES
        for op in ops:
            self._capability &= op.capability

    @staticmethod
    def simplify(ops, neg):
        from .diagonal_operator import DiagonalOperator
        from .scaling_operator import ScalingOperator

        # unpack SumOperators
        opsnew = []
        negnew = []
        for op, ng in zip(ops, neg):
            if isinstance(op, SumOperator):
                opsnew += op._ops
                if ng:
                    negnew += [not n for n in op._neg]
                else:
                    negnew += list(op._neg)
# FIXME: this needs some more work to keep the domain and target unchanged!
#            elif isinstance(op, NullOperator):
#                pass
            else:
                opsnew.append(op)
                negnew.append(ng)
        ops = opsnew
        neg = negnew

        # sort operators according to domains
        sorted = defaultdict(list)
        for op, ng in zip(ops, neg):
            sorted[(op.domain, op.target)].append((op, ng))

        xxops = []
        xxneg = []
        for opset in sorted.values():
            # collect ScalingOperators
            sum = 0.
            opsnew = []
            negnew = []

            dtype = []
            for op, ng in opset:
                if isinstance(op, ScalingOperator):
                    sum += op._factor * (-1 if ng else 1)
                    dtype.append(op._dtype)
                else:
                    opsnew.append(op)
                    negnew.append(ng)
            lastdom = opset[0][0].domain
            del(opset)

            if len(dtype) > 0:
                # Propagate sampling dtypes only if they are all the same
                if all(dtype[0] == ss for ss in dtype):
                    dtype = dtype[0]
                else:
                    dtype = None

            if sum != 0.:
                # try to absorb the factor into a DiagonalOperator
                for i in range(len(opsnew)):
                    if isinstance(opsnew[i], DiagonalOperator):
                        if opsnew[i]._dtype != dtype:
                            continue
                        sum *= (-1 if negnew[i] else 1)
                        opsnew[i] = opsnew[i]._add(sum)
                        sum = 0.
                        break
            if sum != 0 or len(opsnew) == 0:
                # have to add the scaling operator at the end
                opsnew.append(ScalingOperator(lastdom, sum, dtype))
                negnew.append(False)
            del(dtype, sum, lastdom)

            ops = opsnew
            neg = negnew
            # Step 4: combine DiagonalOperators where possible
            processed = [False] * len(ops)
            opsnew = []
            negnew = []
            for i in range(len(ops)):
                if not processed[i]:
                    if isinstance(ops[i], DiagonalOperator):
                        op = ops[i]
                        opneg = neg[i]
                        for j in range(i+1, len(ops)):
                            if isinstance(ops[j], DiagonalOperator) and ops[i]._dtype == ops[j]._dtype:
                                op = op._combine_sum(ops[j], opneg, neg[j])
                                opneg = False
                                processed[j] = True
                        opsnew.append(op)
                        negnew.append(opneg)
                    else:
                        opsnew.append(ops[i])
                        negnew.append(neg[i])
            ops = opsnew
            neg = negnew

            # combine BlockDiagonalOperators where possible
            processed = [False] * len(ops)
            opsnew = []
            negnew = []
            for i in range(len(ops)):
                if not processed[i]:
                    if isinstance(ops[i], BlockDiagonalOperator):
                        op = ops[i]
                        opneg = neg[i]
                        for j in range(i+1, len(ops)):
                            if isinstance(ops[j], BlockDiagonalOperator):
                                op = op._combine_sum(ops[j], opneg, neg[j])
                                opneg = False
                                processed[j] = True
                        opsnew.append(op)
                        negnew.append(opneg)
                    else:
                        opsnew.append(ops[i])
                        negnew.append(neg[i])
            xxops += opsnew
            xxneg += negnew

        dom = domain_union([op.domain for op in xxops])
        tgt = domain_union([op.target for op in xxops])
        return xxops, xxneg, dom, tgt

    @staticmethod
    def make(ops, neg):
        """Build a SumOperator (or something simpler if possible)

        Parameters
        ----------
        ops: list of LinearOperator
            Individual operators of the sum.
        neg: list of bool
            Same length as ops.
            If True then the corresponding operator gets a minus in the sum.
        """
        ops = tuple(ops)
        neg = tuple(neg)
        if len(ops) == 0:
            raise ValueError("ops is empty")
        if len(ops) != len(neg):
            raise ValueError("length mismatch between ops and neg")
        ops, neg, dom, tgt = SumOperator.simplify(ops, neg)
        if len(ops) == 1:
            return -ops[0] if neg[0] else ops[0]
        return SumOperator(ops, neg, dom, tgt, _callingfrommake=True)

    @property
    def adjoint(self):
        return self.make([op.adjoint for op in self._ops], self._neg)

    def apply(self, x, mode):
        self._check_mode(mode)
        res = None
        for op, neg in zip(self._ops, self._neg):
            tmp = op.apply(x.extract(op._dom(mode)), mode)
            if res is None:
                res = -tmp if neg else tmp
            else:
                res = res.flexible_addsub(tmp, neg)
        return res

    def draw_sample(self, from_inverse=False, device_id=-1):
        if from_inverse:
            raise NotImplementedError(
                "cannot draw from inverse of this operator")
        res = None
        for op in self._ops:
            from .simple_linear_operators import NullOperator
            if isinstance(op, NullOperator):
                continue
            tmp = op.draw_sample(from_inverse, device_id)
            res = tmp if res is None else res.unite(tmp)
        return res

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in self._ops)
        return "SumOperator:\n"+indent(subs)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        f = []
        o = []
        for op in self._ops:
            tf, to = op.simplify_for_constant_input(
                c_inp.extract_part(op.domain))
            f.append(tf)
            o.append(to)

        from ..multi_domain import MultiDomain
        if not isinstance(self._target, MultiDomain):
            fullop = None
            for to, n in zip(o, self._neg):
                op = to if not n else -to
                fullop = op if fullop is None else fullop + op
            return None, fullop

        from .simplify_for_const import ConstCollector
        cc = ConstCollector()
        fullop = None
        for tf, to, n in zip(f, o, self._neg):
            cc.add(tf, to.target)
            op = to if not n else -to
            fullop = op if fullop is None else fullop + op
        return cc.constfield, fullop



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from typing import Callable, Optional

import numpy as np
from jax import numpy as jnp
from jax.tree_util import Partial, tree_map

from ..tree_math.vector_math import any as tree_any

exp = partial(tree_map, jnp.exp)
sqrt = partial(tree_map, jnp.sqrt)
log = partial(tree_map, jnp.log)
log1p = partial(tree_map, jnp.log1p)


def _standard_to_laplace(xi, *, alpha):
    from jax.scipy.stats import norm

    norm_logcdf = partial(tree_map, norm.logcdf)

    res = (xi < 0) * (norm_logcdf(xi) + jnp.log(2))
    res -= (xi > 0) * (norm_logcdf(-xi) + jnp.log(2))
    return res * alpha


def laplace_prior(alpha) -> Partial:
    """
    Takes random normal samples and outputs samples distributed according to

    .. math::
        P(x|a) = exp(-|x|/a)/a/2

    """

    return Partial(_standard_to_laplace, alpha=alpha)


def _standard_to_normal(xi, *, mean, std):
    return mean + std * xi


def normal_prior(mean, std) -> Partial:
    """Match standard normally distributed random variables to non-standard
    variables.
    """
    return Partial(_standard_to_normal, mean=mean, std=std)


def _normal_to_standard(y, *, mean, std):
    return (y - mean) / std


def normal_invprior(mean, std) -> Partial:
    """Get the inverse transform to `normal_prior`."""
    return Partial(_normal_to_standard, mean=mean, std=std)


def lognormal_moments(mean, std):
    """Compute the cumulants a log-normal process would need to comply with the
    provided mean and standard-deviation `std`
    """
    if tree_any(mean <= 0.0):
        raise ValueError(f"`mean` must be greater zero; got {mean!r}")
    if tree_any(std <= 0.0):
        raise ValueError(f"`std` must be greater zero; got {std!r}")

    logstd = sqrt(log1p((std / mean) ** 2))
    logmean = log(mean) - 0.5 * logstd**2
    return logmean, logstd


def _standard_to_lognormal(xi, *, log_mean, log_std):
    return exp(_standard_to_normal(xi, mean=log_mean, std=log_std))


def lognormal_prior(mean, std, *, _log_mean=None, _log_std=None) -> Partial:
    """Moment-match standard normally distributed random variables to log-space

    Takes random normal samples and outputs samples distributed according to

    .. math::
        P(xi|mu,sigma) \\propto exp(mu + sigma * xi)

    such that the mean and standard deviation of the distribution matches the
    specified values.
    """
    if _log_mean is None and _log_std is None:
        _log_mean, _log_std = lognormal_moments(mean, std)
    return Partial(_standard_to_lognormal, log_mean=_log_mean, log_std=_log_std)


def _lognormal_to_standard(y, *, log_mean, log_std):
    return _normal_to_standard(log(y), mean=log_mean, std=log_std)


def lognormal_invprior(mean, std, *, _log_mean=None, _log_std=None) -> Partial:
    """Get the inverse transform to `lognormal_prior`."""
    if _log_mean is None and _log_std is None:
        _log_mean, _log_std = lognormal_moments(mean, std)
    return Partial(_lognormal_to_standard, log_mean=_log_mean, log_std=_log_std)


def _standard_to_uniform(xi, *, a_min, scale):
    from jax.scipy.stats import norm

    return a_min + scale * tree_map(norm.cdf, xi)


def uniform_prior(a_min=0.0, a_max=1.0) -> Partial:
    """Transform a standard normal into a uniform distribution.

    Parameters
    ----------
    a_min : float
        Minimum value.
    a_max : float
        Maximum value.
    """
    from jax.scipy.stats import norm

    if (
        isinstance(a_min, float)
        and isinstance(a_max, float)
        and a_min == 0.0
        and a_max == 1.0
    ):
        return Partial(partial(tree_map, norm.cdf))

    scale = a_max - a_min
    return Partial(_standard_to_uniform, a_min=a_min, scale=scale)


def interpolator(
    func: Callable,
    xmin: float,
    xmax: float,
    *,
    step: Optional[float] = None,
    num: Optional[int] = None,
    table_func: Optional[Callable] = None,
    inv_table_func: Optional[Callable] = None,
    return_inverse: Optional[bool] = False,
):  # Adapted from NIFTy
    """
    Evaluate a function point-wise by interpolation.  Can be supplied with a
    table_func to increase the interpolation accuracy, Best results are
    achieved when `lambda x: table_func(func(x))` is roughly linear.

    Parameters
    ----------
    func : function
        Function to interpolate.
    xmin : float
        The smallest value for which `func` will be evaluated.
    xmax : float
        The largest value for which `func` will be evaluated.
    step : float
        Distance between sampling points for linear interpolation. Either of
        `step` or `num` must be specified.
    num : int
        The number of interpolation points. Either of `step` of `num` must be
        specified.
    table_func : function
        Non-linear function applied to the tabulated function in order to
        transform the table to a more linear space.
    inv_table_func : function
        Inverse of `table_func`.
    return_inverse : bool
        Whether to also return the interpolation of the inverse of `func`. Only
        sensible if `func` is invertible.
    """
    # from scipy.interpolate import CubicSpline

    if step is not None and num is not None:
        ve = "either but not both of `step` and `num` must be specified"
        raise ValueError(ve)
    if step is not None:
        # numpy array to enforce evaluating `func` with float64 inputs
        xs = np.arange(xmin, xmax + step, step)
    elif num is not None:
        # numpy array to enforce evaluating `func` with float64 inputs
        xs = np.linspace(xmin, xmax, num)
    else:
        ve = "either of `step` or `num` must be specified"
        raise ValueError(ve)

    ys = func(xs)

    # cast back to jax.numpy after calling `func`
    xs = jnp.array(xs)
    ys = jnp.array(ys)

    if table_func is not None:
        if inv_table_func is None:
            raise ValueError("no `inv_table_func` specified")
        ys = table_func(ys)

    # interpolator = CubicSpline(xs, ys)
    # deriv = interpolator.derivative()

    def interp(x):
        # res = interpolator(x)
        res = jnp.interp(x, xs, ys)
        if inv_table_func is not None:
            res = inv_table_func(res)
        return res

    if return_inverse:

        def inverse_interp(y):
            if table_func is not None:
                y = table_func(y)
            return jnp.interp(y, ys, xs)

        return interp, inverse_interp

    return interp


def invgamma_prior(a, scale, loc=0.0, step=1e-2) -> Callable:
    """Transform a standard normal into an inverse gamma distribution.

    The pdf of the inverse gamma distribution is defined as follows using
    :math:`q` to denote the scale:

    .. math::

        P(x|q, a) = \\frac{q^a}{\\Gamma(a)}x^{-a -1}
        \\exp \\left(-\\frac{q}{x}\\right)

    That means that for large x the pdf falls off like :math:`x^{(-a -1)}`.
    The mean of the pdf is at :math:`q / (a - 1)` if :math:`a > 1`.
    The mode is :math:`q / (a + 1)`.

    This transformation is implemented as a linear interpolation which maps a
    Gaussian onto an inverse gamma distribution.

    Parameters
    ----------
    a : float
        The shape-parameter of the inverse-gamma distribution.
    scale : float
        The scale-parameter of the inverse-gamma distribution.
    loc : float
        An option shift of the whole distribution.
    step : float
        Distance between sampling points for linear interpolation.
    """
    from scipy.stats import invgamma, norm

    if not jnp.isscalar(a) or not jnp.isscalar(loc):
        te = (
            "Shape `a` and location `loc` must be of scalar type"
            f"; got {type(a)} and {type(loc)} respectively"
        )
        raise TypeError(te)
    if loc == 0.0:
        # Pull out `scale` to interpolate less
        s2i = lambda x: invgamma.ppf(norm._cdf(x), a=a)
    elif jnp.isscalar(scale):
        s2i = lambda x: invgamma.ppf(norm._cdf(x), a=a, loc=loc, scale=scale)
    else:
        raise TypeError("`scale` may only be array-like for `loc == 0.`")

    xmin, xmax = -8.2, 8.2  # (1. - norm.cdf(8.2)) * 2 < 1e-15
    standard_to_invgamma_interp = interpolator(
        s2i, xmin, xmax, step=step, table_func=jnp.log, inv_table_func=jnp.exp
    )

    def standard_to_invgamma(x):
        # Allow for array-like `scale` without separate interpolations and only
        # interpolate for shape `a` and `loc`
        if loc == 0.0:
            return standard_to_invgamma_interp(x) * scale
        return standard_to_invgamma_interp(x)

    return standard_to_invgamma


def invgamma_invprior(a, scale, loc=0.0, step=1e-2) -> Callable:
    """Get the inverse transformation to `invgamma_prior`."""
    from scipy.stats import invgamma, norm

    xmin, xmax = -8.2, 8.2  # (1. - norm.cdf(8.2)) * 2 < 1e-15
    _, invgamma_to_standard = interpolator(
        lambda x: invgamma.ppf(norm._cdf(x), a=a, loc=loc, scale=scale),
        xmin,
        xmax,
        step=step,
        table_func=jnp.log,
        inv_table_func=jnp.exp,
        return_inverse=True,
    )
    return invgamma_to_standard



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..domains.unstructured_domain import UnstructuredDomain
from ..multi_domain import MultiDomain
from ..sugar import makeField
from .linear_operator import LinearOperator


class Multifield2Vector(LinearOperator):
    """Flatten a MultiField and return a Field with unstructured domain and the
    same number of degrees of freedom.

    Parameters
    ----------
    domain: MultiDomain
        Domain of the operator

    Notes
    -----
    This operator works only on MultiFields that have the same dtype for all entries.
    """

    def __init__(self, domain):
        if not isinstance(domain, MultiDomain):
            raise NotImplementedError("This operator only works on MultiDomains")
        self._dof = domain.size
        self._domain = domain
        self._target = DomainTuple.make(UnstructuredDomain(self._dof))
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        ii = 0
        if mode == self.TIMES:
            dtype = _unique_dtype([vv.dtype for vv in x.values()])
            res = np.empty_like(next(iter(x.val.values())), shape=self.target.shape, dtype=dtype)
            for key in self.domain.keys():
                arr = x.val[key].flatten()
                res[ii:ii + arr.size] = arr
                ii += arr.size
        else:
            res = {}
            for key in self.domain.keys():
                n = self.domain[key].size
                shp = self.domain[key].shape
                res[key] = x.val[ii:ii + n].reshape(shp)
                ii += n
        return makeField(self._tgt(mode), res)


def _unique_dtype(lst):
    if all(ll == lst[0] for ll in lst):
        return lst[0]
    raise RuntimeError("Dtype is not unique")
